/*
 *			    Squirt main common
 *			    ------------------
 *
 *		Jean II '00
 *
 * Handle all the common code for the various e-squirt applications.
 * Main calls to handle e-Squirt...
 *
 * This file is released under the GPL license.
 */

/*
Copyright (C) 1995-2001 Hewlett-Packard Company 

This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation; either version 2 of the License, or (at your option) any later version. 

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details. 

You should have received a copy of the GNU General Public License along with this program; if not, write to the Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. 
*/

/***************************** INCLUDES *****************************/

#include "squirt.h"
#include "squirt_common.h"
#include "squirt_xml.h"
#include "squirt_socket.h"

#ifdef HAS_ULTRA
#include "squirt_ultra.h"
#endif /* HAS_ULTRA */

#ifdef HAS_OBEX
#include "squirt_obex_irda.h"
#endif /* HAS_OBEX */

#ifdef HAS_BT
#include "squirt_obex_bt.h"
#endif /* HAS_BT */

/************************** SQUIRT SENDER **************************/
/*
 * This is the main entry point of the library to send E-Squirts and Beacons.
 * We don't do much, we just look at which transport is requested and
 * call the approriate function...
 */

/* ---------------------------------------------------------------- */
/*
 * Send a e-Squirt
 * Depending on the options, we try the relevant protocols and transports
 *
 * Returns :
 *	-1 : error
 *	 0 : success
 */
int
send_squirt(es_squirt *	content,	/* Payload */
	    es_opt *	options,	/* How to process */
	    es_trans *	extra,		/* Optional Transport data */
	    gi_obget *	reply,		/* Optional returned Obex headers */
	    gi_flags *	flags)		/* Optional Rx flags */
{
  int			ret = -1;	/* Fail */

  /* Create XML string if needed (with relevant options) */
  if(options->do_xml != OPT_NO_XML)
    {
      ret = create_xml(content, options);
      if(ret)
	goto cleanup;
    }

  /* Check if regular socket or Obex is requested. May fail... */
  if((options->transport & OPT_IRDA) || (options->transport == OPT_UNDEFINED))
    {
#ifdef HAS_OBEX
      /* Check if regular OBEX is requested. */
      if((options->protocol & OPT_OBEX) ||
	 (options->protocol == OPT_UNDEFINED))
	{
	  /* Just do it (Obex options will be dealt in there) */
	  ret = send_irda_obex(content, options, extra, reply, flags,
			       OPT_OBEX_NO_STDIAS);
	  /* Success ? */
	  if(ret >= 0)
	    goto cleanup;
	  /* Do we want to try the plain OBEX ? */
	  if(options->obex.ias == OPT_OBEX_DO_STDIAS)
	    {
	      /* Try again connecting to "OBEX" */
	      ret = send_irda_obex(content, options, extra, reply, flags,
				   OPT_OBEX_DO_STDIAS);
	      /* Success ? */
	      if(ret >= 0)
		goto cleanup;
	    }
	}
      /* Check if OBEX GET is requested (squirt_get). */
      if(options->protocol & OPT_OBEX_GET)
	{
	  /* Just do it (Obex options will be dealt in there) */
	  ret = send_irda_obex_get(content, options, extra, reply, flags,
				   OPT_OBEX_NO_STDIAS);
	  /* Success ? */
	  if(ret >= 0)
	    goto cleanup;
	  /* Do we want to try the plain OBEX ? */
	  if(options->obex.ias == OPT_OBEX_DO_STDIAS)
	    {
	      /* Try again connecting to "OBEX" */
	      ret = send_irda_obex_get(content, options, extra, reply, flags,
				       OPT_OBEX_DO_STDIAS);
	      /* Success ? */
	      if(ret >= 0)
		goto cleanup;
	    }
	}
#endif /* HAS_OBEX */
      /* Check if regular IrDA socket is requested. */
      if((options->protocol & OPT_RAW) ||
	 (options->protocol == OPT_UNDEFINED))
	{
	  /* Check if it is a squirt. */
	  if((options->model == OPT_SQUIRT) ||
	     (options->model == OPT_UNDEFINED))
	    {
	      /* Just do it */
	      ret = send_socket(content, options, extra, OPT_SQUIRT);
	      /* Success ? */
	      if(ret >= 0)
		goto cleanup;
	    }
	  /* Check if it is a beacon. */
	  if((options->model == OPT_BEACON) ||
	     (options->model == OPT_UNDEFINED))
	    {
	      /* Just do it */
	      ret = send_socket(content, options, extra, OPT_BEACON);
	      /* Success ? */
	      if(ret >= 0)
		goto cleanup;
	    }
	}
    }

#ifdef HAS_BT
  /* Check if regular BlueTooth is requested. May fail... */
  if((options->transport & OPT_BLUETOOTH) ||
     (options->transport == OPT_UNDEFINED))
    {
      /* Check if regular OBEX is requested. */
      if((options->protocol & OPT_OBEX) ||
	 (options->protocol == OPT_UNDEFINED))
	{
	  /* Just do it (Obex options will be dealt in there) */
	  ret = send_bt_obex(content, options, extra, reply, flags);
	  /* Success ? */
	  if(ret >= 0)
	    goto cleanup;
	}
      /* Check if OBEX GET is requested (squirt_get). */
      if(options->protocol & OPT_OBEX_GET)
	{
	  /* Just do it (Obex options will be dealt in there) */
	  ret = send_bt_obex_get(content, options, extra, reply, flags);
	  /* Success ? */
	  if(ret >= 0)
	    goto cleanup;
	}
    }
#endif /* HAS_BT */

  /* Important note : if the caller use OPT_UNDEFINED, if both socket and
   * Obex connect did fail AND if the other side has an active IrDA stack,
   * THEN the Ultra packet below will never be sent.
   * The reason is that Linux will have the LAP layer connected to the other
   * side, and can't send Ultra packets in this mode (just dropped).
   *
   * On the other hand, if the caller use OPT_UNDEFINED, if both socket and
   * Obex connect did fail AND if the other side does NOT have an active
   * IrDA stack (i.e. support Ultra only), things will work fine.
   */

#ifdef HAS_ULTRA
  /* Check if Ultra is requested. Ultra never fails */
  if((options->transport & OPT_ULTRA) ||
     (options->transport == OPT_UNDEFINED))
    {
      /* Just do it (Obex option will be dealt in there) */
      ret = send_ultra(content, options, extra);
    }
#endif /* HAS_ULTRA */

 cleanup:
  if(content->cleanup == OPT_DATA_DO_FREE)
    {
      free(content->data);
      content->data = NULL;
#ifdef DEBUG_ALLOC
      fprintf(stderr, "%s : free(content->data)\n", __FUNCTION__);
#endif
    }

  return(ret);
}

/************************* SQUIRT_RECEIVER *************************/
/*
 * This is the main entry point of the library to receive
 * E-Squirts and Beacons.
 * We don't do much, we just look at which transport is requested and
 * call the approriate function...
 */

/* ---------------------------------------------------------------- */
/*
 * Create all the listeners and add them in descriptor list
 */
es_listen *
prepare_listen(int *		plast,	/* Pointer to last fd value */
	       es_opt *		options)
{
  es_listen *	list = NULL;	/* List of listener */
  es_listen *	current;	/* Current listener */
  int		last_fd = -1;	/* Last file descriptor */

  /* Check if regular socket or Obex is requested. */
  if((options->transport & OPT_IRDA) || (options->transport == OPT_UNDEFINED))
    {
#ifdef HAS_OBEX
      /* Check if regular OBEX is requested. */
      if((options->protocol & OPT_OBEX) ||
	 (options->protocol & OPT_OBEX_GET) ||
	 (options->protocol == OPT_UNDEFINED))
	{
	  /* Try to add an squirt-Obex listener */
	  current = add_irda_obex(list, options, OPT_OBEX_NO_STDIAS);
	  if(current != NULL)
	    {
	      list = current;			/* New listener list */
	      if(current->sock > last_fd)	/* Update last_fd */
		last_fd = current->sock;
	    }
	  /* Do we want to receive on the plain OBEX ? */
	  if(options->obex.ias == OPT_OBEX_DO_STDIAS)
	    {
	      /* Try to add an squirt-Obex listener */
	      current = add_irda_obex(list, options, OPT_OBEX_DO_STDIAS);
	      if(current != NULL)
		{
		  list = current;		/* New listener list */
		  if(current->sock > last_fd)	/* Update last_fd */
		    last_fd = current->sock;
		}
	    }
	}
#endif /* HAS_OBEX */

      /* Check if regular IrDA socket is requested. */
      if((options->protocol & OPT_RAW) || (options->protocol == OPT_UNDEFINED))
	{
	  /* Check if squirt is requested. */
	  if((options->model == OPT_SQUIRT) ||
	     (options->model == OPT_UNDEFINED))
	    {
	      /* Try to add an squirt-socket listener */
	      current = add_socket(list, options, OPT_SQUIRT);
	      if(current != NULL)
		{
		  list = current;		/* New listener list */
		  if(current->sock > last_fd)	/* Update last_fd */
		    last_fd = current->sock;
		}
	    }
	  /* Check if beacon is requested. */
	  if((options->model == OPT_BEACON) ||
	     (options->model == OPT_UNDEFINED))
	    {
	      /* Try to add an beacon-socket listener */
	      current = add_socket(list, options, OPT_BEACON);
	      if(current != NULL)
		{
		  list = current;		/* New listener list */
		  if(current->sock > last_fd)	/* Update last_fd */
		    last_fd = current->sock;
		}
	    }
	}
    }

#ifdef HAS_BT
  /* Check if BlueTooth is requested. */
  if((options->transport & OPT_BLUETOOTH) ||
     (options->transport == OPT_UNDEFINED))
    {
      /* Check if regular OBEX is requested. */
      if((options->protocol & OPT_OBEX) ||
	 (options->protocol & OPT_OBEX_GET) ||
	 (options->protocol == OPT_UNDEFINED))
	{
	  /* Try to add an squirt-Obex listener */
	  current = add_bt_obex(list, options);
	  if(current != NULL)
	    {
	      list = current;			/* New listener list */
	      if(current->sock > last_fd)	/* Update last_fd */
		last_fd = current->sock;
	    }
	}
    }
#endif /* HAS_BT */

#ifdef HAS_ULTRA
  /* Check if Ultra is requested. */
  if((options->transport & OPT_ULTRA) ||
     (options->transport == OPT_UNDEFINED))
    {
      /* Check if Ultra-Obex is requested */
      if((options->protocol & OPT_OBEX) ||
	 (options->protocol == OPT_UNDEFINED))
	{
	  /* Try to add an Ultra-Obex listener */
	  current = add_ultra(list, options, UPID_OBEX);
	  if(current != NULL)
	    {
	      list = current;			/* New listener list */
	      if(current->sock > last_fd)	/* Update last_fd */
		last_fd = current->sock;
	    }
	}
      /* Check if raw Ultra is requested  */
      if((options->protocol & OPT_RAW) ||
	 (options->protocol == OPT_UNDEFINED))
	{
	  /* Try to add an raw Ultra listener */
	  current = add_ultra(list, options, UPID_BEACON);
	  if(current != NULL)
	    {
	      list = current;			/* New listener list */
	      if(current->sock > last_fd)	/* Update last_fd */
		last_fd = current->sock;
	    }
	  /* Backward compatibility with old beacons */
	  current = add_ultra(list, options, UPID_BEACON_OLD);
	  if(current != NULL)
	    {
	      list = current;			/* New listener list */
	      if(current->sock > last_fd)	/* Update last_fd */
		last_fd = current->sock;
	    }
	}

    }
#endif /* HAS_ULTRA */

  /* Get back to the caller */
  *plast = last_fd;
  return(list);
}

/* ---------------------------------------------------------------- */
/*
 * Close all listener and free pieces...
 */
int
cleanup_listen(es_listen *		list)
{
  es_listen *		current;	/* Position in list */
  es_listen *		temp;		/* Node to remove */

  /* Go through the listener list and close every socket */
  current = list;
  while(current != NULL)
    {
      /* Unlink from list */
      temp = current;
      current = current->next;
      /* Call the destroy handler (will free(temp))*/
      temp->destroy(temp);
    }
  return(0);
}

/* ---------------------------------------------------------------- */
/*
 * After the accept, process the incomming data and call the user.
 * Returns what the user callback returns
 */
int
call_rcv_callback(es_rx_data *	rx_data,
		  gi_obget *	reply,		/* Headers we want back */
		  gi_flags *	flags)		/* Corresponding flags */
{
  int		ret = 0;

  /* Check if the user has asked to filter out Squirts or Beacons */
  if(rx_data->listen_options->model != OPT_UNDEFINED)
    /* Check if it is the right kind - won't work for Ultra-Raw */
    if(rx_data->options->model != rx_data->listen_options->model)
      {
	fprintf(stderr, "Wrong model - discarding...\n");
	ret = -1;
      }

  /* Parse the XML payload (if needed) */
  if((ret >= 0) && (rx_data->listen_options->do_xml != OPT_NO_XML))
    ret = parse_xml(rx_data->content, rx_data->options);

  /* Call the user callback if it exists */
  if(ret >= 0)
    {
      if(rx_data->callback != (squirt_handler) NULL)
	{
	  /* Do it */
	  ret = rx_data->callback(rx_data->content, rx_data->options,
				  rx_data->extra, reply, flags);
	  /* Cleanup if user has not already done so */
	  if((rx_data->content->data != NULL) &&
	     (rx_data->content->cleanup == OPT_DATA_DO_FREE))
	    {
	      free(rx_data->content->data);
	      rx_data->content->data = NULL;
#ifdef DEBUG_ALLOC
	      fprintf(stderr, "%s: free(content->data)\n", __FUNCTION__);
#endif
	    }
	}
      else
	ret = 0;	/* No callback, but it's ok... */
    }
  else
    {
      /* Either wrong model or failed XML parsing - cleanup */
      if((rx_data->content->data != NULL) &&
	 (rx_data->content->cleanup == OPT_DATA_DO_FREE))
	{
	  free(rx_data->content->data);
	  rx_data->content->data = NULL;
#ifdef DEBUG_ALLOC
	  fprintf(stderr, "%s: free(content->data)\n", __FUNCTION__);
#endif
	}
    }

  return(ret);
}

/* ---------------------------------------------------------------- */
/*
 * After the accept, process the incomming data and call the user.
 * Note on the return value :
 *	-1 == error
 *	 0 == ok, callback not called (so the user has not got the data yet)
 *	 1 == ok, callback called (the user has already a copy of the data)
 */
int
finish_receive(es_listen *	current,
	       es_rx_data *	rx_data)
{
  int		ret = 0;

  /* Call the post processing handler (read data and process it) */
  /* Note : this may call call_rcv_callback() */
  ret = current->post_handler(current, rx_data);

  /* Close the accept socket if it exist */
  current->end_recv(current);

  /* Finish the post processing and call user callback */
  if(ret == 0)
    {
      ret = call_rcv_callback(rx_data, NULL, NULL);
      if(ret < 0)
	ret = -1;
      else
	ret = (rx_data->callback != (squirt_handler) NULL);
    }

  return(ret);
}

/* ---------------------------------------------------------------- */
/*
 * Handle e-Squirt and Beacon reception.
 * Can be used for a single reception if callback is NULL, or can
 * continously monitor the sockets and call the callback (which avoid
 * to close and re-open them and to miss data).
 * We create and close the sockets here.
 *
 * Note on the content and option parameters : we will use the same
 * physical structure for each reception. If the callback is synchronous
 * (no fork), he will finish with them before returning to us. If the
 * callback fork, the forked version will have its own personal copy...
 * On the other hand, callback using threads may want to make a copy
 * for the forked thread...
 *
 * Returns :
 *	-1 : error
 *	 0 : receive synchronous squirt successful
 *	 1 : user callback asked for termination
 */
int
recv_squirt(es_squirt *		content,	/* Rx payload */
	    es_opt *		options,	/* How to listen */
	    es_trans *		extra,		/* Transport data */
	    squirt_handler	callback,	/* Rx Callback */
	    es_server_data *	control)	/* Control server behaviour */
{
  es_listen *		list;		/* List of listener */
  es_listen *		current;	/* Current listener */
  fd_set		rfds;		/* File descriptors for select */
  int			last_fd;	/* Last fd */
  struct timeval	tv;		/* Select timeout */
  es_opt		listen_options;	/* Listening options */
  es_server_data	null_control;	/* Just if control is NULL */
  es_rx_data		rx_data;	/* Receiver data */
  es_post		post_data;	/* Post processing data */
  int			ret;

  /* Check the control stuff */
  if(control == NULL)
    {
      memset((char *) &null_control, 0, sizeof(es_server_data));
      control = &null_control;
    }

  /* Add all the listeners */
  memcpy((char *) &listen_options, (char *) options, sizeof(es_opt));
  list = prepare_listen(&last_fd, &listen_options);
  if(list == NULL)
    return(-1);

  /* Set the Rx Data, we'll need that later ;-) */
  rx_data.content = content;
  rx_data.options = options;
  rx_data.extra = extra;
  rx_data.callback = callback;
  rx_data.listen_options = &listen_options;

  /* Receive packets */
  while(1)
    {
      /* Reset the receive data structures so that we don't carry
       * over data from previous receptions */
      rx_data.content->label[0] = '\0';
      rx_data.content->url[0] = '\0';
      rx_data.content->type[0] = '\0';
      memset(rx_data.options, 0, sizeof(es_opt));
      /* Don't reset extra because of userdata and
       * because controlled by options */

      /* Guess what ? We must re-generate rdfs each time */
      FD_ZERO(&rfds);
      current = list;
      while(current != NULL)
	{
	  FD_SET(current->sock, &rfds);	/* Add to descriptor list */
	  current = current->next;
	}

      /* Init timeout value */
      tv.tv_sec = control->timeout;
      tv.tv_usec = 0;

      /* Wait until something happens */
      ret = select(last_fd + 1, &rfds, NULL, NULL,
		   (control->timeout) ? &tv : NULL);

      /* Check if there was an error or timeout (0=timeout, -1=signal) */
      if(ret <= 0)
	{
	  /* Any select handler ? */
	  if(control->timeback != (select_handler) NULL)
	    {
	      /* Call it */
	      ret = control->timeback(ret, control->cbdata);
	      if(!ret)
		continue;
	    }
	  fprintf(stderr, "Unhandled signal - exiting...\n");
	  break;
	}

      /* Go through the listener list and check every socket */
      current = list;
      while(current != NULL)
	{
	  /* Check if it was this socket */
	  if((FD_ISSET(current->sock, &rfds)))
	    break;
	  current = current->next;
	}

      /* Check if we have one socket */
      if(current == NULL)
	{
	  fprintf(stderr, "Unknown interuption...\n");
	  continue;
	}

      /* Call the handler to accept the connection */
      ret = current->pre_handler(current, &rx_data);
      if(ret < 0)
	continue;

      /* Is there any accept handler ? */
      if(control->acceptback != (accept_handler) NULL)
	{
	  /* Set up post processing data (default : no fork, no m/thread) */
	  post_data.param1 = list;
	  post_data.param2 = current;
	  post_data.do_post = 1;		/* Do recv */
	  post_data.do_cleanup = 0;		/* Don't close sockets */

	  /* Call the user handler (will call the post processing handler) */
	  ret = control->acceptback(&post_data, &rx_data, control->cbdata);
	  if(ret)
	    break;
	}
      else
	{
	  /* No accept handler : call the post handler ourselves */
	  ret = finish_receive(current, &rx_data);
	  if(ret == 0)
	    break;
	}
    }

  /* Cleanup all the sockets */
  cleanup_listen(list);

  return(ret);
}

/************************** SERVER CONTROL **************************/
/*
 * All the uglyness for people who want to control the behaviour
 * of the main server loop and handle forking/threaded model.
 */

/* ---------------------------------------------------------------- */
/*
 * Update post processing data after a fork()
 * Use only for forking model
 *
 * After a fork, the data is already duplicated in each process space.
 * So, we just set up proper flags...
 */
int
post_fork_squirt(es_post *	post_data,
		 es_rx_data *	rx_data,
		 int		fpid)		/* Return code of the fork */
{
  /* Are we in the parent ? */
  if(fpid != 0)
    {
      /* No post processing in the main server */
      post_data->do_post = 0;
    }
  else
    {
      /* We are in the child */
      /* In the child, close properly all sockets */
      post_data->do_cleanup = 1;
    }

  return(0);
}

/* ---------------------------------------------------------------- */
/*
 * Duplicate post processing data before spawning a new thread.
 * Use only for multi-threading model
 *
 * Threads share the same process space, so each thread need a copy...
 * Note : this function has *NOT* been tested.
 */
int
pre_thread_squirt(es_post *	old_post_data,
		  es_post *	new_post_data,
		  es_rx_data *	old_rx_data,
		  es_rx_data *	new_rx_data,
		  es_squirt *	new_content,	/* New Payload */
		  es_opt *	new_options,	/* New Options */
		  es_trans *	new_extra)	/* New Extra */
{
  es_listen *		new_desc;
  es_listen *		old_desc = old_post_data->param2;

  /* Allocate some memory for the receiver description */
  new_desc = (es_listen *) malloc(sizeof(es_listen));
  if(new_desc == NULL)
    {
      free(new_post_data);
      fprintf(stderr, "Malloc failed\n");
      return(-1);
    }
  /* Fill the new fake listener and detach from the old one */
  memcpy(new_desc, old_desc, sizeof(es_listen));
  new_desc->sock = -1;
  old_desc->rx_sock = -1;
  new_desc->handle = (void *) NULL;
  old_desc->rx_handle = (void *) NULL;
  new_desc->next = (es_listen *) NULL;	/* End of list */

  /* Put the new post processing descriptor in order */
  new_post_data->param1 = new_desc;
  new_post_data->param2 = new_desc;
  /* Do post processing in the child */
  new_post_data->do_post = 1;
  /* Cleanup new_desc in the child (not the full socket list) */
  new_post_data->do_cleanup = 1;

  /* No post processing in the main server */
  old_post_data->do_post = 0;
  /* In fact, the main server won't do any post processing at all, so
   * the caller need to call post_process_squirt() only on new_post_data */

  /* Set up the new rx_data */
  memcpy(new_rx_data, old_rx_data, sizeof(es_rx_data));
  /* And put the new payload in the old one */
  old_rx_data->content = new_content;
  old_rx_data->options = new_options;
  old_rx_data->extra = new_extra;

  return(0);
}

/* ---------------------------------------------------------------- */
/*
 * Do the post processing for reception. Called by the user callback.
 */
int
post_process_squirt(es_post *		post_data,
		    es_rx_data *	rx_data)
{
  es_listen *		list;		/* List of listener */
  es_listen *		current;	/* Current listener */
  int			ret = 0;

  /* Set variables */
  if(post_data == NULL)
    return(-1);
  list    = post_data->param1;
  current = post_data->param2;
  if((current == NULL) || (list == NULL))
    return(-2);

  /* Do the post processing (if needed) */
  if(post_data->do_post)
    ret = finish_receive(current, rx_data);

  /* Close all the sockets (listeners and connected - if needed) */
  if(post_data->do_cleanup)
    cleanup_listen(list);

  return(ret);
}

/************************** GET INTERFACE **************************/
/*
 * The Get-Interface service look like a cut down dump HTTP service.
 * The client can request various files, and the server handle
 * the incomming request.
 *
 * This service is currently only available only on IrDA OBEX
 * (using the GET primitive).
 *
 * Note that this service is somewhat mixed with the squirt+get
 * functionality available througth the normal send_squirt() path.
 */

/* ---------------------------------------------------------------- */
/*
 * Issue a get-interface command and return the result, if possible...
 */
int
get_interface(gi_obget *	request,	/* What to get */
	      gi_flags *	flags,		/* How to get it */
	      es_trans *	extra)		/* Transport data */
{
  int			ret = -1;	/* Fail */

  /* Check if regular socket or Obex is requested. May fail... */
  if((flags->transport & OPT_IRDA) || (flags->transport == OPT_UNDEFINED))
    {
#ifdef HAS_OBEX
      /* Check if regular OBEX is requested. */
      if((flags->protocol & OPT_OBEX) ||
	 (flags->protocol & OPT_OBEX_GET) || 
	 (flags->protocol == OPT_UNDEFINED))
	{
	  /* Just do it (Obex flags will be dealt in there) */
	  ret = get_interface_irda_obex(request, flags, extra,
					OPT_OBEX_NO_STDIAS);
	  /* Success ? */
	  if(ret >= 0)
	    return(ret);
	  /* Do we want to try the plain OBEX ? */
	  if(flags->obex.ias == OPT_OBEX_DO_STDIAS)
	    {
	      /* Try again connecting to "OBEX" */
	      ret = get_interface_irda_obex(request, flags, extra,
					    OPT_OBEX_DO_STDIAS);
	      /* Success ? */
	      if(ret >= 0)
		return(ret);
	    }
	}
#endif /* HAS_OBEX */
    }

#ifdef HAS_BT
  /* Check if regular socket or Obex is requested. May fail... */
  if((flags->transport & OPT_BLUETOOTH) ||
     (flags->transport == OPT_UNDEFINED))
    {
      /* Check if regular OBEX is requested. */
      if((flags->protocol & OPT_OBEX) ||
	 (flags->protocol & OPT_OBEX_GET) || 
	 (flags->protocol == OPT_UNDEFINED))
	{
	  /* Just do it (Obex flags will be dealt in there) */
	  ret = get_interface_bt_obex(request, flags, extra);
	  /* Success ? */
	  if(ret >= 0)
	    return(ret);
	}
    }
#endif /* HAS_BT */

  return(ret);
}

/* ---------------------------------------------------------------- */
/*
 * Set the get interface handler
 * Here, we only set the callback. This callback will only be used
 * if the user start an Obex listener (server via recv_squirt()),
 * and then for each incomming GET request this callback will be called.
 * Note : there is no flags because we are fully dependant on the options
 * passed to recv_squirt().
 */
int
set_obget_handler(obget_handler	getback,
		  void *	userdata)
{
  int ret = -1;

  /* Go through all of them... */
#ifdef HAS_OBEX
  ret = set_obget_handler_obex(getback, userdata);
#endif /* HAS_OBEX */

  return(ret);
}
