/*
 *			    Squirt BlueTooth Obex common
 *			    ----------------------------
 *
 *		Jean II '00
 *
 * Handle all the common BlueTooth Obex code for the various e-squirt
 * applications.
 *
 * This file is released under the GPL license.
 */

/*
Copyright (C) 1995-2001 Hewlett-Packard Company 

This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation; either version 2 of the License, or (at your option) any later version. 

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details. 

You should have received a copy of the GNU General Public License along with this program; if not, write to the Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. 
*/

/***************************** INCLUDES *****************************/

/* Header order is *very* important */
#include "squirt_common.h"

#include <bluetooth.h>
#include <hci.h>		/* BlueTooth HCI lib */
#include <hci_lib.h>
#include <sdp.h>		/* BlueTooth SDP library */
#include <sdp_lib.h>
#include <rfcomm.h>		/* struct sockaddr_rc */

#include "squirt_obex.p.h"

/************************ CONSTANTS & MACROS ************************/

/* Defined in squirt_common.h */
extern const unsigned char UUGLI_SQUIRT_TARGET[];

/* Name of the SDP service we use */
#define SQUIRT_SERVICE_NAME	"hp:esquirt"

/* RfCOMM/L2CAP channel used */
/* We should think of dynamic allocation and using SDP */
#define BT_CHANNEL	19

/* Number of devices returned by Inquiry */
#define BT_INQ_NUM	20

/* Memory allocation for discovery */
#define DISC_MAX_DEVICES 10
#define DISC_BUF_LEN	sizeof(struct irda_device_list) + \
			sizeof(struct irda_device_info) * DISC_MAX_DEVICES

/************************* RANDOM UTILITIES *************************/
/*
 * Various utilities that don't fall in a specific category...
 */

/* ---------------------------------------------------------------- */
/*
 * Map the field of various structures to point to each other
 */
static inline void
map_obex_struct_fields(es_squirt *	content,
		       es_trans *	extra,
		       es_obex *	obex)
{
  memset(obex, 0, sizeof(es_obex));

  obex->payload = content->data;
  obex->size = content->datalen;
  if(extra != NULL)
    {
      obex->filename = extra->filename;
      obex->usertype = extra->type;
    }
  obex->transport = OPT_BLUETOOTH;
}

/************************** BLUETOOTH SDP **************************/

/* 1) Search on connection (local/remote)
 * 2) Set on connection (local)
 * 3) Search and set 
 */

/* ------------------------------------------------------------------- */
/*
 * Query an existing SDP connection (local ro remote) for the
 * ESquirt service.
 */
static int
sdp_find_channel(sdp_session_t *	conn,	/* SDP connection */
		 struct sockaddr_rc *	rfcomm)	/* Channel found */
{
  uuid_t	service_obex;
  sdp_list_t *	attrid = NULL;
  sdp_list_t *	search = NULL;
  sdp_list_t *	seq = NULL;
  uint32_t	range = 0x0000ffff;
  int		status = -1;

  /* Create UUIDs of desired services */
  /* We ask directly on our UUID, so we don't have to filter
   * records afterward... */
  sdp_uuid128_create(&service_obex, (uint128_t *) UUGLI_SQUIRT_TARGET);
  //sdp_uuid16_create(&service_obex, OBEX_UUID);
  /* Create list of services to search */
  search = sdp_list_append(search, &service_obex);
  /* List of attributes for each service */
  // We should fix that to minimise the amount of data exchanged
  // We should request only our attribs
  attrid = sdp_list_append(NULL, &range);
  /* Get data from remote SDP server */
  status = sdp_service_search_attr_req(conn, search, SDP_ATTR_REQ_RANGE,
				       attrid, &seq);
  if(status || (seq == NULL))
    {
#ifdef DEBUG_BT
      fprintf(stderr, "SDP Service Search failed (status = %d)\n", status);
#endif
      return(-2);
    }

  /* Look in all services found */
  while(seq != NULL)
    {
      sdp_record_t *	svcrec = (sdp_record_t *) seq->data;
      sdp_list_t *	next = seq->next;
      sdp_list_t *	protos;
      int		channel;
      int		ret;

      /* Now, we need to extract the RfCOMM channel */
      ret = sdp_get_access_protos(svcrec, &protos);
      if(ret == 0)
	{
	  channel = sdp_get_proto_port(protos, RFCOMM_UUID);
	  if(channel > 0)
	    {
	      /* Got it ! */
	      rfcomm->rc_channel = channel;
#ifdef DEBUG_BT
	      fprintf(stderr, "RfCOMM channel -> %d\n", channel);
#endif
	      //sdp_list_free(protos, 0);
	      sdp_list_free(seq, 0);
	      return(0);
	    }
	  //sdp_list_free(protos, 0);
	}

      /* Cleanup */
      sdp_list_remove(seq, svcrec);
      free(svcrec);
      seq = next;
    }

  /* Not found */
  return(-1);;
}

/* ------------------------------------------------------------------- */
/*
 * Set a SDP record for the ESquirt service on the local SDP server
 */
static int
sdp_set_record(sdp_session_t *		conn,	/* SDP connection */
	       struct sockaddr_rc *	rfcomm)	/* Channel found */
{
  sdp_list_t *	svclass_id = NULL;
  //sdp_list_t *	pfseq = NULL;
  sdp_list_t *	apseq = NULL;
  sdp_list_t *	root = NULL;
  uuid_t	root_uuid;
  uuid_t	svc_uuid;
  uuid_t	l2cap_uuid;
  uuid_t	rfcomm_uuid;
  uuid_t	obex_uuid;
  //sdp_profile_desc_t profile[1];
  sdp_list_t *	aproto;
  sdp_list_t *	proto[3];
  sdp_record_t	svcrec;
  uint8_t	u8 = rfcomm->rc_channel;
  sdp_data_t *	channel;
  int		status;

  /* Initialise service record */
  memset((void *)&svcrec, 0, sizeof(sdp_record_t));
  svcrec.handle = 0xffffffff;

  /* Record will be at top level */
  sdp_uuid16_create(&root_uuid, PUBLIC_BROWSE_GROUP);
  root = sdp_list_append(0, &root_uuid);
  sdp_set_browse_groups(&svcrec, root);

  /* Provide E-Squirt service */
  sdp_uuid128_create(&svc_uuid, (uint128_t *) UUGLI_SQUIRT_TARGET);
  svclass_id = sdp_list_append(0, &svc_uuid);
  sdp_set_service_classes(&svcrec, svclass_id);

  /* We don't conform to any profile */
  //sdp_uuid16_create(&profile[0].uuid, XXX_PROFILE_ID);
  //profile[0].version = 0x0100;
  //pfseq = sdp_list_append(0, profile);
  //sdp_set_profile_descs(&svcrec, pfseq);

  /* Use L2CAP protocol */
  sdp_uuid16_create(&l2cap_uuid, L2CAP_UUID);
  proto[0] = sdp_list_append(0, &l2cap_uuid);
  apseq = sdp_list_append(0, proto[0]);

  /* And RfCOMM protocol */
  sdp_uuid16_create(&rfcomm_uuid, RFCOMM_UUID);
  proto[1] = sdp_list_append(0, &rfcomm_uuid);
  channel = sdp_data_alloc(SDP_UINT8, &u8);
  proto[1] = sdp_list_append(proto[1], channel);
  apseq = sdp_list_append(apseq, proto[1]);

  /* And finally OBEX protocol */
  sdp_uuid16_create(&obex_uuid, OBEX_UUID);
  proto[2] = sdp_list_append(0, &obex_uuid);
  apseq = sdp_list_append(apseq, proto[2]);

  /* Add protocol list */
  aproto = sdp_list_append(0, apseq);
  sdp_set_access_protos(&svcrec, aproto);

  /* Cleanup */
  sdp_data_free(channel);
  sdp_list_free(proto[0], 0);
  sdp_list_free(proto[1], 0);
  sdp_list_free(proto[2], 0);
  sdp_list_free(apseq, 0);
  sdp_list_free(aproto, 0);

  /* User readable comment */
  /* Actually, we will use that to identify the service */
  sdp_set_info_attr(&svcrec, SQUIRT_SERVICE_NAME, 0, 0);

  /* Push record on the server */
  status = sdp_record_register(conn, &svcrec, SDP_RECORD_PERSIST);
  if(status)
    {
      fprintf(stderr, "Service Record registration failed.\n");
      return(-1);
    }

#ifdef DEBUG_BT
  fprintf(stderr, "E-Squirt HP service registered\n");
#endif
  return(0);
}

/* ------------------------------------------------------------------- */
/*
 * Get a SDP channel.
 * After this call, rfcomm.rc_channel contains the proper channel
 * number.
 * We will Add a SDP record description for the OBEX service we provide
 * if needed, or reuse the old one.
 */
static int
sdp_get_channel(struct sockaddr_rc *	rfcomm)
{
  sdp_session_t *	conn;
  int			ret;

  /* Connect to the local SDP server */
  conn = sdp_connect(BDADDR_ANY, BDADDR_LOCAL, 0);
  if(conn == NULL)
    {
      printf("Failed to connect to local SDP server\n");
      return(-1);
    }

  /* Check if the service already exist */
  ret = sdp_find_channel(conn, rfcomm);

  /* If doesn't already exist, create it */
  if(ret < 0)
    {
      /* Pick a channel - may want to be dynamic */
      rfcomm->rc_channel = BT_CHANNEL;

      /* Create a record */
      ret = sdp_set_record(conn, rfcomm);
    }

  sdp_close(conn);
  return(ret);
}

/* ------------------------------------------------------------------- */
/*
 * Query a peer to find the RfCOMM channel it is using
 */
static int
sdp_query_peer(struct sockaddr_rc *	rfcomm)
{
  sdp_session_t *	conn;
  int			ret;

  /* Connect to the local SDP server */
  conn = sdp_connect(BDADDR_ANY, &rfcomm->rc_bdaddr, 0);
  if(conn == NULL)
    {
#ifdef DEBUG_BT
      fprintf(stderr, "Failed to connect to remote SDP server\n");
#endif
      return(-1);
    }

  /* Check if the service already exist */
  ret = sdp_find_channel(conn, rfcomm);

  sdp_close(conn);
  return(ret);
}

/************************* OBEX PUT SENDER *************************/
/*
 * Sending an object over Obex...
 */

/* ---------------------------------------------------------------- */
/*
 * Discover the server address...
 */
static int
bt_obex_discover(struct sockaddr_rc *	rfcomm)
{
  inquiry_info		ii[BT_INQ_NUM];
  int			n;
  char			str[20];
  inquiry_info *	info = ii;
  int			i;
  int			ret;

  /* Perform an Inquiry */
  n = hci_inquiry(-1,		/* Any HCI interface */
		  10,		/* 10*1.28 seconds */
		  BT_INQ_NUM,
		  NULL,
		  &info,
		  0);
  if(n <= 0)
    {
      if(n < 0)
	fprintf(stderr, "BT Inquiry failed\n");
#ifdef DEBUG_BT
      else
	fprintf(stderr, "BT Inquiry found nothing\n");
#endif
      return(-1);
    }

  /* Check all peers, pick the first one that support E-Squirt */
  for(i = 0; i < n; i++)
    {
      /* Quick hack */
      info = &ii[i];
      ba2str(&info->bdaddr, str);
#ifdef DEBUG_BT
      fprintf(stderr, "Found BT peer %s !\n", str);
#endif

      /* Try to find the proper SDP record */
      bacpy(&rfcomm->rc_bdaddr, &info->bdaddr);
      ret = sdp_query_peer(rfcomm);
      if(ret == 0)
	{
#ifdef DEBUG_BT
	  fprintf(stderr, "Found E-Squirt on BT peer %s !\n", str);
#endif
	  /* Got it ! */
	  return(0);
	}
    }

  /* Not found */
#ifdef DEBUG_BT
  fprintf(stderr, "No BT peer support E-Squirt !\n");
#endif
  return(-1);
}

/* ---------------------------------------------------------------- */
/*
 * Send a squirt to the server...
 */
int
send_bt_obex(es_squirt *	content,	/* Payload */
	     es_opt *		options,	/* How to process */
	     es_trans *		extra,		/* Optional Transport data */
	     gi_obget *		reply,		/* Optional Rx Obex headers */
	     gi_flags *		flags)		/* Optional Rx flags */
{
  obex_t *		handle = NULL;	/* Obex connection handle */
  obex_object_t *	object;		/* Obex object to send */
  es_obexsend		cbdata;		/* Data for our callback... */
  es_obex		command;	/* The PUT command to create */
  struct sockaddr_rc	rfcomm;
  int			ret = 0;
  int			get_ret = -1;	/* Return of the get command */

  /* Build command descr */
  map_obex_struct_fields(content, extra, &command);

  /* Create a handle for this connection */
  handle = OBEX_Init(OBEX_TRANS_BLUETOOTH, obex_client_event, 0);
  if(handle == (obex_t *) NULL)
    {
      fprintf(stderr, "OBEX_Init failed\n");
      return(-1);
    }

  /* Set as user data of the Obex handle */
  cbdata.handle = handle;
  cbdata.status = OBEX_STAT_CONT;
  cbdata.put = 0;
  cbdata.reply = reply;
  cbdata.flags = flags;
  cbdata.extra = extra;
  cbdata.options = options;
  cbdata.has_replied = 0;
  OBEX_SetUserData(handle, (void *) &cbdata);

  /* Increase throughput */
  OBEX_SetTransportMTU(handle, OBEX_MAXIMUM_MTU, OBEX_MAXIMUM_MTU);

  /* Find a peer */
  if((options->linkaddr == OPT_DO_LINKADDR) && (extra != NULL))
    {
      /* Peer specified */
#ifdef DEBUG_BT
      fprintf(stderr, "Trying link address [%s]\n", extra->linkaddr);
#endif
      str2ba(extra->linkaddr, &rfcomm.rc_bdaddr);

      /* Find the channel on the peer */
      ret = sdp_query_peer(&rfcomm);
      if(ret < 0)
	{
	  fprintf(stderr, "BT peer %s doesn't support E-Squirt !\n",
		  extra->linkaddr);
	  return(ret);
	}
    }
  else
    {
      /* No peer specified, find one */
      ret = bt_obex_discover(&rfcomm);
      if(ret < 0)
	return(ret);
      /* Save address, so that next time the caller don't have to do
       * an Inquiry ;-) */
      ba2str(&rfcomm.rc_bdaddr, extra->linkaddr);
      options->linkaddr = OPT_DO_LINKADDR;
    }

  /* Connect to the peer */
  ret = BtOBEX_TransportConnect(handle, BDADDR_ANY,
				&rfcomm.rc_bdaddr, rfcomm.rc_channel);
  if(ret < 0)
    {
      OBEX_Cleanup(handle);
      return(ret);
    }

  /* Creating a sesssion is optional. It improves performance for long
   * transaction and allow to do get_interface, but increases latency */
  if(options->obex.session != OPT_OBEX_NO_SESSION)
    {
      /* Start the Obex session (optional - may want to add count hdr) */
      object = OBEX_ObjectNew(handle, OBEX_CMD_CONNECT);
      ret = obex_send_object(&cbdata, object);
      if(ret != OBEX_STAT_SENT)
	{
	  OBEX_Cleanup(handle);
	  return(-1);
	}
    }

  /* Build Obex PUT object containing the e-Squirt */
  object = OBEX_ObjectNew(handle, OBEX_CMD_PUT);
  create_obex_headers(handle, object, &command, options);

  /* Process the "put" (send it) */
  ret = obex_send_object(&cbdata, object);

  /* Sesssion is optional. */
  if(options->obex.session != OPT_OBEX_NO_SESSION)
    {
      /* Check if we need to get an interface at this point */
      if(cbdata.put && cbdata.has_replied &&
	 (options->get_iface == OPT_DO_GIFACE))
	{
	  /* Do a get-interface in the present session */
	  get_ret = simple_get_interface(handle, reply, flags, extra);

	  /* Reset stuff back to properly finish the session */
	  OBEX_SetUserCallBack(handle, obex_client_event, &cbdata);
	}

      /* Finish the Obex session */
      object = OBEX_ObjectNew(handle, OBEX_CMD_DISCONNECT);
      ret = obex_send_object(&cbdata, object);
      /* We don't really care if this one fails */
    }

  /* The end */
  OBEX_Cleanup(handle);
  if(cbdata.put < 0)
    ret = cbdata.put;		/* Return Obex failure code */
  else
    ret = -1 + (cbdata.put > 0) + (cbdata.has_replied > 0) + (get_ret >= 0);
  return(ret);
}

/************************* OBEX GET SENDER *************************/

/* ---------------------------------------------------------------- */
/*
 * Issue a get-interface command and return the result, if possible...
 */
int
get_interface_bt_obex(gi_obget *	request,	/* What to get */
		      gi_flags *	flags,		/* How to get it */
		      es_trans *	extra)		/* Transport data */
{
  obex_t *		handle = NULL;	/* Obex connection handle */
  obex_object_t *	object;		/* Obex object to send */
  gi_obgetsend		cbdata;		/* Data for our callback... */
  struct sockaddr_rc	rfcomm;
  int			ret = 0;
  int			get_ret = -1;	/* Return of the get command */

  /* Create a handle for this connection */
  handle = OBEX_Init(OBEX_TRANS_BLUETOOTH, obget_client_event, 0);
  if(handle == (obex_t *) NULL)
    {
      fprintf(stderr, "OBEX_Init failed\n");
      return(-1);
    }

  /* Set as user data of the Obex handle */
  cbdata.handle = handle;
  cbdata.status = OBEX_STAT_CONT;
  cbdata.get = 0;
  cbdata.request = request;
  cbdata.flags = flags;
  cbdata.extra = extra;
  OBEX_SetUserData(handle, (void *) &cbdata);

  /* Increase throughput */
  OBEX_SetTransportMTU(handle, OBEX_MAXIMUM_MTU, OBEX_MAXIMUM_MTU);

  /* Find a peer */
  if((flags->linkaddr == OPT_DO_LINKADDR) && (extra != NULL))
    {
      /* Peer specified */
#ifdef DEBUG_BT
      fprintf(stderr, "Trying link address [%s]\n", extra->linkaddr);
#endif
      str2ba(extra->linkaddr, &rfcomm.rc_bdaddr);

      /* Find the channel on the peer */
      ret = sdp_query_peer(&rfcomm);
      if(ret < 0)
	{
	  fprintf(stderr, "BT peer %s doesn't support E-Squirt !\n",
		  extra->linkaddr);
	  return(ret);
	}
    }
  else
    {
      /* No peer specified, find one */
      ret = bt_obex_discover(&rfcomm);
      if(ret < 0)
	return(ret);
      /* Save address, so that next time the caller don't have to do
       * an Inquiry ;-) */
      ba2str(&rfcomm.rc_bdaddr, extra->linkaddr);
      flags->linkaddr = OPT_DO_LINKADDR;
    }

  /* Connect to the peer */
  ret = BtOBEX_TransportConnect(handle, BDADDR_ANY,
				&rfcomm.rc_bdaddr, rfcomm.rc_channel);
  if(ret < 0)
    {
      OBEX_Cleanup(handle);
      return(ret);
    }

  /* Creating a sesssion is optional. It improves performance for long
   * transaction and allow to do get_interface, but increases latency */
  if(flags->obex.session != OPT_OBEX_NO_SESSION)
    {
      /* Start the Obex session (optional : may want to add count hdr) */
      object = OBEX_ObjectNew(handle, OBEX_CMD_CONNECT);
      ret = obget_send_object(&cbdata, object);
      if(ret != OBEX_STAT_SENT)
	{
	  OBEX_Cleanup(handle);
	  return(-1);
	}
    }

  /* Build Obex GET object containing the request parameters */
  object = OBEX_ObjectNew(handle, OBEX_CMD_GET);
  create_obget_headers(handle, object, request, flags, extra);

  /* Process the "get" (send it and get answer back) */
  get_ret = obget_send_object(&cbdata, object);

  /* Sesssion is optional. */
  if(flags->obex.session != OPT_OBEX_NO_SESSION)
    {
      /* Finish the Obex session */
      object = OBEX_ObjectNew(handle, OBEX_CMD_DISCONNECT);
      ret = obget_send_object(&cbdata, object);
      /* We don't really care if this one fails */
    }

  /* The end - The data will be in the request struct, if any */
  OBEX_Cleanup(handle);
  if(cbdata.get < 0)
    return(cbdata.get);
  if(get_ret != OBEX_STAT_SENT)
    return(-3);
  if(cbdata.get == 0)
    return(-2);
  return(0);
}

/* ---------------------------------------------------------------- */
/*
 * Send a squirt over a GET, Issue a get-interface command and
 * return the result, if possible...
 */
int
send_bt_obex_get(es_squirt *	squirt,		/* Optional squirt */
		 es_opt *	options,	/* For the squirt */
		 es_trans *	extra,		/* Transport data */
		 gi_obget *	request,	/* What to get */
		 gi_flags *	flags)		/* How to get it */
{
  obex_t *		handle = NULL;	/* Obex connection handle */
  obex_object_t *	object;		/* Obex object to send */
  gi_obgetsend		cbdata;		/* Data for our callback... */
  es_obex		command;	/* The optional PUT command */
  struct sockaddr_rc	rfcomm;
  int			ret = 0;
  int			get_ret = -1;	/* Return of the get command */

  map_obex_struct_fields(squirt, extra, &command);

  /* Create a handle for this connection */
  handle = OBEX_Init(OBEX_TRANS_BLUETOOTH, obget_client_event, 0);
  if(handle == (obex_t *) NULL)
    {
      fprintf(stderr, "OBEX_Init failed\n");
      return(-1);
    }

  /* Set as user data of the Obex handle */
  cbdata.handle = handle;
  cbdata.status = OBEX_STAT_CONT;
  cbdata.get = 0;
  cbdata.request = request;
  cbdata.flags = flags;
  cbdata.extra = extra;
  OBEX_SetUserData(handle, (void *) &cbdata);

  /* Increase throughput */
  OBEX_SetTransportMTU(handle, OBEX_MAXIMUM_MTU, OBEX_MAXIMUM_MTU);

  /* Find a peer */
  if((flags->linkaddr == OPT_DO_LINKADDR) && (extra != NULL))
    {
      /* Peer specified */
#ifdef DEBUG_BT
      fprintf(stderr, "Trying link address [%s]\n", extra->linkaddr);
#endif
      str2ba(extra->linkaddr, &rfcomm.rc_bdaddr);

      /* Find the channel on the peer */
      ret = sdp_query_peer(&rfcomm);
      if(ret < 0)
	{
	  fprintf(stderr, "BT peer %s doesn't support E-Squirt !\n",
		  extra->linkaddr);
	  return(ret);
	}
    }
  else
    {
      /* No peer specified, find one */
      ret = bt_obex_discover(&rfcomm);
      if(ret < 0)
	return(ret);
      /* Save address, so that next time the caller don't have to do
       * an Inquiry ;-) */
      ba2str(&rfcomm.rc_bdaddr, extra->linkaddr);
      flags->linkaddr = OPT_DO_LINKADDR;
    }

  /* Connect to the peer */
  ret = BtOBEX_TransportConnect(handle, BDADDR_ANY,
				&rfcomm.rc_bdaddr, rfcomm.rc_channel);
  if(ret < 0)
    {
      OBEX_Cleanup(handle);
      return(ret);
    }

  /* Creating a sesssion is optional. It improves performance for long
   * transaction and allow to do get_interface, but increases latency */
  if(flags->obex.session != OPT_OBEX_NO_SESSION)
    {
      /* Start the Obex session (optional - may want to add count hdr) */
      object = OBEX_ObjectNew(handle, OBEX_CMD_CONNECT);
      ret = obget_send_object(&cbdata, object);
      if(ret != OBEX_STAT_SENT)
	{
	  OBEX_Cleanup(handle);
	  return(-1);
	}
    }

  /* Those headers are reserved for the Get-Interface options */
  options->obex.name = OPT_OBEX_NO_NAME;
  if(flags->obex.target == OPT_OBEX_TARGET_USER)
    options->model = OPT_UNDEFINED;
  /* If we put a Squirt in the body, we can't put some request data as well
   * so disable it */
  if((flags->squirt_get != OPT_NO_SQUIRT_GET) &&
     (request->datalen >= 0) && (request->data != NULL))
    {
      fprintf(stderr, "Can't process request->data with a Squirt - dropped\n");
      free(request->data);
      request->data = NULL;
#ifdef DEBUG_ALLOC
      fprintf(stderr, "%s: free(request->data)\n", __FUNCTION__);
#endif
    }
  /* No type has a special meaning in GET requests, it means name overflow,
   * so don't generate it in a squirt ! */
  if(options->obex.type == OPT_OBEX_TYPE_NONE)
    options->obex.type = OPT_UNDEFINED;
  /* And if you use options->obex.type which is not REFS, REFX or REFL
   * are definitely asking for troubles... */

  /* Build Obex GET object containing the request parameters */
  object = OBEX_ObjectNew(handle, OBEX_CMD_GET);
  /* Add the Get-Interface headers */
  create_obget_headers(handle, object, request, flags, extra);
  /* Add the squirt header+payload (if not disabled) */
  if(flags->squirt_get != OPT_NO_SQUIRT_GET)
    create_obex_headers(handle, object, &command, options);

  /* Process the "get" (send it) */
  get_ret = obget_send_object(&cbdata, object);

  /* Sesssion is optional. */
  if(flags->obex.session != OPT_OBEX_NO_SESSION)
    {
      /* Finish the Obex session */
      object = OBEX_ObjectNew(handle, OBEX_CMD_DISCONNECT);
      ret = obget_send_object(&cbdata, object);
      /* We don't really care if this one fails */
    }

  /* The end - The data will be in the request struct, if any */
  OBEX_Cleanup(handle);
  if(cbdata.get < 0)
    return(cbdata.get);
  if(get_ret != OBEX_STAT_SENT)
    return(-3);
  if(cbdata.get == 0)
    return(-2);
  return(0);
}

/************************** OBEX RECEIVER **************************/
/*
 * This is where we actually receive beacons and e-Squirt
 * over the IrObex transport (regular Obex over TinyTP).
 * Under Linux, we simply use the excellent Obex support ;-)
 */

/* ---------------------------------------------------------------- */
/*
 * Handle the reception of some data on the Obex server
 * Just create a separate Obex handle via OBEX_ServerAccept()
 *
 * Note on Obex user data : as opposed to the send, the user data is
 * valid only within each individual function. It is therefore *mandatory*
 * to set it prior to any call to OBEX_HandleInput().
 */
int
bt_obex_accept(es_listen *	desc,
	       es_rx_data *	rx_data)
{
  es_obexrecv		cbdata;		/* Data for our callback... */
  int			ret;

  /* Set our stuff as user data of the Obex handle */
  cbdata.handle = desc->handle;
  cbdata.status = OBEX_STAT_CONT;
  OBEX_SetUserData(desc->handle, (void *) &cbdata);

  /* Get answer from Obex lib */
  while(cbdata.status == OBEX_STAT_CONT)
    {
      ret = OBEX_HandleInput(desc->handle, 1);
      if(ret < 0)
	return(ret);
    }
  /* Check if we got what we want */
  if(cbdata.status != OBEX_STAT_ACCEPT)
    {
      fprintf(stderr, "Invalid status 0x%X\n", cbdata.status);
      return(-1);
    }

  /* Reset handle data back to default */
  OBEX_SetUserData(desc->handle, (void *) NULL);

  /* Get a new handle for the new connection */
  desc->rx_handle = OBEX_ServerAccept(desc->handle, obex_recv_event, NULL);

  return(0);
}

/* ---------------------------------------------------------------- */
/*
 * Post processing handler for Obex socket...
 * Receive and parse the data appropriately...
 */
int
bt_obex_recv(es_listen *	desc,
	     es_rx_data *	rx_data)
{
  es_squirt *		content = rx_data->content;
  es_opt *		options = rx_data->options;
  es_trans *		extra   = rx_data->extra;
  es_obexrecv		cbdata;		/* Data for our callback... */
  int			ret;

  /* Set our stuff as user data of the Obex handle */
  cbdata.handle = desc->rx_handle;
  cbdata.status = OBEX_STAT_CONT;
  cbdata.put = -1;
  cbdata.get = -1;
  cbdata.session = OPT_OBEX_NO_SESSION;
  map_obex_struct_fields(content, extra, &(cbdata.data));
  cbdata.options = options;
  cbdata.userdata = rx_data;
  OBEX_SetUserData(desc->rx_handle, (void *) &cbdata);

  /* Set some options */
  options->transport = OPT_BLUETOOTH;
  options->protocol = OPT_OBEX;
  options->model = OPT_UNDEFINED;
  options->obex.target = OPT_UNDEFINED;
  options->obex.ias = OPT_UNDEFINED;
  options->obex.name = OPT_OBEX_NO_NAME;
  options->obex.type = OPT_OBEX_TYPE_NONE;
  options->obex.session = OPT_OBEX_NO_SESSION;

  if(extra != NULL)
    {
      int			sock;
      struct sockaddr_rc	rfcomm;
      int			len;

      /* Get the connected socket */
      sock = OBEX_GetFD(desc->rx_handle);

      /* Get out info about our socket (remote peer) */
      len = sizeof(struct sockaddr_rc);
      if(getpeername(sock, (struct sockaddr*) &rfcomm, &len))
	perror("getsockname");
      else
	{
	  ba2str(&rfcomm.rc_bdaddr, extra->linkaddr);
	  options->linkaddr = OPT_DO_LINKADDR;
#ifdef DEBUG_BT
	  fprintf(stderr, "Receive from BT peer %s / %d!\n",
		 extra->linkaddr, rfcomm.rc_channel);
#endif
	}
    }

  /* Get answer from Obex lib */
  while(cbdata.status == OBEX_STAT_CONT)
    {
      ret = OBEX_HandleInput(desc->rx_handle, 1);
      if(ret < 0)
	break;
    }

  /* Note : at this point, both Obex and XML should be properly
   * decapsulated. See process_put_request() for details...
   * In most cases, we even have called the user callback... */
  return(cbdata.put);	/* What finish_receive() expects */
}

/* ---------------------------------------------------------------- */
/*
 * Destroy the Obex receiver. Cleanup its handle.
 */
int
bt_obex_end_recv(es_listen *	desc)
{
  /* Close the handle */
  if(desc->rx_handle != NULL)
    OBEX_Cleanup(desc->rx_handle);
  desc->rx_handle = 0;
  return(0);
}

/* ---------------------------------------------------------------- */
/*
 * Destroy a Obex listener. Cleanup the handle and free pieces.
 */
int
bt_obex_destroy(es_listen *	desc)
{
  /* Close the handles */
  if(desc->handle != NULL)
    OBEX_Cleanup(desc->handle);
  if(desc->rx_handle != NULL)
    OBEX_Cleanup(desc->rx_handle);
  /* Free pieces */
  desc->next = (es_listen *) NULL;
  free(desc);
  return(0);
}

/* ---------------------------------------------------------------- */
/*
 * Add a receiver description for Bt socket to the list
 */
es_listen *
add_bt_obex(es_listen *	list,
	    es_opt *	options)
{
  es_listen *		desc;
  obex_t *		handle = NULL;	/* Obex connection handle */
  int			sock;		/* Socket itself */
  struct sockaddr_rc	rfcomm;
  int			ret;

  /* Add a record to SDP and get channel number */
  ret = sdp_get_channel(&rfcomm);
  if(ret < 0)
    {
      fprintf(stderr, "SDP registration failed (no SDP server ?)\n");
      return(NULL);
    }

  /* Create a handle for this connection */
  handle = OBEX_Init(OBEX_TRANS_BLUETOOTH, obex_server_event,
		     OBEX_FL_KEEPSERVER);
  if(handle == (obex_t *) NULL)
    {
      fprintf(stderr, "OBEX_Init failed\n");
      return(NULL);
    }

  /* Increase throughput */
  OBEX_SetTransportMTU(handle, OBEX_MAXIMUM_MTU, OBEX_MAXIMUM_MTU);

  /* Start the Obex server */
  ret = BtOBEX_ServerRegister(handle, BDADDR_ANY, rfcomm.rc_channel);
  if(ret < 0)
    {
      fprintf(stderr, "OBEX : Unable to register the socket (no Bt stack ?)\n");
      return(NULL);
    }

  /* Need to get the socket here !!! */
  sock = OBEX_GetFD(handle);

  /* Allocate some memory for the receiver description */
  desc = (es_listen *) malloc(sizeof(es_listen));
  if(desc == NULL)
    {
      fprintf(stderr, "Malloc failed\n");
      return(NULL);
    }

  /* Fill the description */
  desc->sock = sock;
  desc->rx_sock = -1;
  desc->ultra = 0;
  desc->handle = (void *) handle;
  desc->rx_handle = (void *) NULL;
  desc->pre_handler = &bt_obex_accept;
  desc->post_handler = &bt_obex_recv;
  desc->end_recv = &bt_obex_end_recv;
  desc->destroy = &bt_obex_destroy;

  /* Link to the list and goodbye */
  desc->next = list;
  return(desc);
}
