/*
 *			    Squirt IrDA Obex common
 *			    -----------------------
 *
 *		Jean II '00
 *
 * Handle all the common IrDA Obex code for the various e-squirt applications.
 *
 * This file is released under the GPL license.
 */

/*
Copyright (C) 1995-2001 Hewlett-Packard Company 

This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation; either version 2 of the License, or (at your option) any later version. 

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details. 

You should have received a copy of the GNU General Public License along with this program; if not, write to the Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. 
*/

#include "squirt_obex.p.h"

/************************ CONSTANTS & MACROS ************************/

/* Memory allocation for discovery */
#define DISC_MAX_DEVICES 10
#define DISC_BUF_LEN	sizeof(struct irda_device_list) + \
			sizeof(struct irda_device_info) * DISC_MAX_DEVICES

/************************* RANDOM UTILITIES *************************/
/*
 * Various utilities that don't fall in a specific category...
 */

/* ---------------------------------------------------------------- */
/*
 * Map the field of various structures to point to each other
 */
static inline void
map_obex_struct_fields(es_squirt *	content,
		       es_trans *	extra,
		       es_obex *	obex)
{
  memset(obex, 0, sizeof(es_obex));

  obex->payload = content->data;
  obex->size = content->datalen;
  if(extra != NULL)
    {
      obex->filename = extra->filename;
      obex->usertype = extra->type;
    }
  obex->transport = OPT_IRDA;
}

/************************* OBEX PUT SENDER *************************/
/*
 * Sending an object over Obex...
 */

/* ---------------------------------------------------------------- */
/*
 * Discover the server address...
 * Simplified version of discovery in squirt_socket.c
 */
static int
irda_obex_discover(char *	nickname,
		   int		filter)
{
  int			fd;			/* An IrDA socket */
  struct irda_device_list *	list;		/* List of device */
  unsigned char		buf[DISC_BUF_LEN];	/* Actual memory allocation */
  int			len;
  unsigned char		hints[4];		/* Hint bits we filter on */
  int			i;

  /* Get some IrDA socket */
  fd = socket(AF_IRDA, SOCK_STREAM, 0);
  if(fd < 0)
    {
      perror("Creating IrDA socket (no IrDA stack ?) ");
      return(-1);
    }

  /* Do we need to filter the log ? */
  if(filter)
    {
      /* We want only devices that advertise OBEX hint */
      hints[0] = HINT_EXTENSION;
      hints[1] = HINT_OBEX;

      if(setsockopt(fd, SOL_IRLMP, IRLMP_HINT_MASK_SET, hints, sizeof(hints))) 
	{
	  close(fd);
	  return(-1);
	}
    }

  /* Set the list to point to the correct place */
  list = (struct irda_device_list *) buf;
  len = DISC_BUF_LEN;

  /* Ask for the discovery log */
  if(getsockopt(fd, SOL_IRLMP, IRLMP_ENUMDEVICES, buf, &len))
    {
      /* Discovery log empty (normal case) or error */
      printf("No device in sight !\n");
      close(fd);
      return(-1);
    }
  /* Finished with it */
  close(fd);

  /* Is there any addresses ? */
  if(list->len <= 0)
    /* Discovery log empty (exceptional case) */
    return(-1);

  /* Check based on nickname */
  if(nickname != NULL)
    {
      /* Go through the list */
      for (i=0; i < list->len; i++)
	{
	  /* Try to match the nickname */
	  if(!strcmp(list->dev[i].info, nickname))
	    /* Matched ! -> Return the address */
	    return(list->dev[i].daddr);
	}
      /* No match */
      return(-1);
    }

  /* Let's do it *really* simple */
  return(list->dev[0].daddr);
}

/* ---------------------------------------------------------------- */
/*
 * Send a squirt to the server...
 */
int
send_irda_obex(es_squirt *	content,	/* Payload */
	       es_opt *		options,	/* How to process */
	       es_trans *	extra,		/* Optional Transport data */
	       gi_obget *	reply,		/* Optional Rx Obex headers */
	       gi_flags *	flags,		/* Optional Rx flags */
	       int		opt_ias)
{
  obex_t *		handle = NULL;	/* Obex connection handle */
  obex_object_t *	object;		/* Obex object to send */
  es_obexsend		cbdata;		/* Data for our callback... */
  es_obex		command;	/* The PUT command to create */
  int			obex_flags;
  int			ret = 0;
  int			get_ret = -1;	/* Return of the get command */

  /* Build command descr */
  map_obex_struct_fields(content, extra, &command);

  /* Create a handle for this connection */
  if(options->filter == OPT_NO_FILTER)
    obex_flags = 0;
  else
    obex_flags = OBEX_FL_FILTERHINT;
  handle = OBEX_Init(OBEX_TRANS_IRDA, obex_client_event, obex_flags);
  if(handle == (obex_t *) NULL)
    {
      fprintf(stderr, "OBEX_Init failed\n");
      return(-1);
    }

  /* Set as user data of the Obex handle */
  cbdata.handle = handle;
  cbdata.status = OBEX_STAT_CONT;
  cbdata.put = 0;
  cbdata.reply = reply;
  cbdata.flags = flags;
  cbdata.extra = extra;
  cbdata.options = options;
  cbdata.has_replied = 0;
  OBEX_SetUserData(handle, (void *) &cbdata);

  /* Increase throughput */
  OBEX_SetTransportMTU(handle, OBEX_IRDA_OPT_MTU, OBEX_IRDA_OPT_MTU);

  if((options->linkaddr == OPT_DO_LINKADDR) && (extra != NULL))
    {
      /* Use this version if you want to specify the IrDA destination
       * address of the Obex peer. Jean II */
      int			daddr;		/* Destination address */
      struct sockaddr_irda	peer;

      /* Discover an address */
      daddr = irda_obex_discover(extra->linkaddr,
				 options->filter == OPT_DO_FILTER);
      if(daddr == -1)
	return(-1);

      /* Fill in the details of the peer and the service */
      peer.sir_family = AF_IRDA;
      peer.sir_lsap_sel = LSAP_ANY;
      peer.sir_addr = daddr;
      if(opt_ias == OPT_OBEX_DO_STDIAS)
	strcpy(peer.sir_name, OBEX_STD_SERVICE_NAME);
      else
	strcpy(peer.sir_name, OBEX_SERVICE_NAME);

      /* Connect (socket) to the peer */
      ret = OBEX_TransportConnect(handle,
				  (struct sockaddr *) &peer,
				  sizeof(peer));
      if(ret < 0)
	{
	  OBEX_Cleanup(handle);
	  return(ret);
	}
    }
  else
    {
      /* Let's OpenObex do all the hard work for us */
      /* Find a peer, and connect (socket) to it */
      if(opt_ias == OPT_OBEX_DO_STDIAS)
	ret = IrOBEX_TransportConnect(handle, OBEX_STD_SERVICE_NAME);
      else
	ret = IrOBEX_TransportConnect(handle, OBEX_SERVICE_NAME);
      if(ret < 0)
	{
	  OBEX_Cleanup(handle);
	  return(ret);
	}
    }

  /* Creating a sesssion is optional. It improves performance for long
   * transaction and allow to do get_interface, but increases latency */
  if(options->obex.session != OPT_OBEX_NO_SESSION)
    {
      /* Start the Obex session (optional - may want to add count hdr) */
      object = OBEX_ObjectNew(handle, OBEX_CMD_CONNECT);
      ret = obex_send_object(&cbdata, object);
      if(ret != OBEX_STAT_SENT)
	{
	  OBEX_Cleanup(handle);
	  return(-1);
	}
    }

  /* Build Obex PUT object containing the e-Squirt */
  object = OBEX_ObjectNew(handle, OBEX_CMD_PUT);
  create_obex_headers(handle, object, &command, options);

  /* Process the "put" (send it) */
  ret = obex_send_object(&cbdata, object);

  /* Sesssion is optional. */
  if(options->obex.session != OPT_OBEX_NO_SESSION)
    {
      /* Check if we need to get an interface at this point */
      if(cbdata.put && cbdata.has_replied &&
	 (options->get_iface == OPT_DO_GIFACE))
	{
	  /* Do a get-interface in the present session */
	  get_ret = simple_get_interface(handle, reply, flags, extra);

	  /* Reset stuff back to properly finish the session */
	  OBEX_SetUserCallBack(handle, obex_client_event, &cbdata);
	}

      /* Finish the Obex session */
      object = OBEX_ObjectNew(handle, OBEX_CMD_DISCONNECT);
      ret = obex_send_object(&cbdata, object);
      /* We don't really care if this one fails */
    }

  /* The end */
  OBEX_Cleanup(handle);
  if(cbdata.put < 0)
    ret = cbdata.put;		/* Return Obex failure code */
  else
    ret = -1 + (cbdata.put > 0) + (cbdata.has_replied > 0) + (get_ret >= 0);
  return(ret);
}

/************************* OBEX GET SENDER *************************/

/* ---------------------------------------------------------------- */
/*
 * Issue a get-interface command and return the result, if possible...
 */
int
get_interface_irda_obex(gi_obget *	request,	/* What to get */
			gi_flags *	flags,		/* How to get it */
			es_trans *	extra,		/* Transport data */
			int		opt_ias)
{
  obex_t *		handle = NULL;	/* Obex connection handle */
  obex_object_t *	object;		/* Obex object to send */
  gi_obgetsend		cbdata;		/* Data for our callback... */
  int			obex_flags;
  int			ret = 0;
  int			get_ret = -1;	/* Return of the get command */

  /* Create a handle for this connection */
  if(flags->filter == OPT_NO_FILTER)
    obex_flags = 0;
  else
    obex_flags = OBEX_FL_FILTERHINT;
  handle = OBEX_Init(OBEX_TRANS_IRDA, obget_client_event, obex_flags);
  if(handle == (obex_t *) NULL)
    {
      fprintf(stderr, "OBEX_Init failed\n");
      return(-1);
    }

  /* Set as user data of the Obex handle */
  cbdata.handle = handle;
  cbdata.status = OBEX_STAT_CONT;
  cbdata.get = 0;
  cbdata.request = request;
  cbdata.flags = flags;
  cbdata.extra = extra;
  OBEX_SetUserData(handle, (void *) &cbdata);

  /* Increase throughput */
  OBEX_SetTransportMTU(handle, OBEX_IRDA_OPT_MTU, OBEX_IRDA_OPT_MTU);

  if((flags->linkaddr == OPT_DO_LINKADDR) && (extra != NULL))
    {
      /* Use this version if you want to specify the IrDA destination
       * address of the Obex peer. Jean II */
      int			daddr;		/* Destination address */
      struct sockaddr_irda	peer;

      /* Discover an address */
      daddr = irda_obex_discover(extra->linkaddr,
				 flags->filter == OPT_DO_FILTER);
      if(daddr == -1)
	return(-1);

      /* Fill in the details of the peer and the service */
      peer.sir_family = AF_IRDA;
      peer.sir_lsap_sel = LSAP_ANY;
      peer.sir_addr = daddr;
      if(opt_ias == OPT_OBEX_DO_STDIAS)
	strcpy(peer.sir_name, OBEX_STD_SERVICE_NAME);
      else
	strcpy(peer.sir_name, OBEX_SERVICE_NAME);

      /* Connect (socket) to the peer */
      ret = OBEX_TransportConnect(handle,
				  (struct sockaddr *) &peer,
				  sizeof(peer));
      if(ret < 0)
	{
	  OBEX_Cleanup(handle);
	  return(ret);
	}
    }
  else
    {
      /* Let's OpenObex do all the hard work for us */
      /* Find a peer, and connect (socket) to it */
      if(opt_ias == OPT_OBEX_DO_STDIAS)
	ret = IrOBEX_TransportConnect(handle, OBEX_STD_SERVICE_NAME);
      else
	ret = IrOBEX_TransportConnect(handle, OBEX_SERVICE_NAME);
      if(ret < 0)
	{
	  OBEX_Cleanup(handle);
	  return(ret);
	}
    }

  /* Creating a sesssion is optional. It improves performance for long
   * transaction and allow to do get_interface, but increases latency */
  if(flags->obex.session != OPT_OBEX_NO_SESSION)
    {
      /* Start the Obex session (optional : may want to add count hdr) */
      object = OBEX_ObjectNew(handle, OBEX_CMD_CONNECT);
      ret = obget_send_object(&cbdata, object);
      if(ret != OBEX_STAT_SENT)
	{
	  OBEX_Cleanup(handle);
	  return(-1);
	}
    }

  /* Build Obex GET object containing the request parameters */
  object = OBEX_ObjectNew(handle, OBEX_CMD_GET);
  create_obget_headers(handle, object, request, flags, extra);

  /* Process the "get" (send it and get answer back) */
  get_ret = obget_send_object(&cbdata, object);

  /* Sesssion is optional. */
  if(flags->obex.session != OPT_OBEX_NO_SESSION)
    {
      /* Finish the Obex session */
      object = OBEX_ObjectNew(handle, OBEX_CMD_DISCONNECT);
      ret = obget_send_object(&cbdata, object);
      /* We don't really care if this one fails */
    }

  /* The end - The data will be in the request struct, if any */
  OBEX_Cleanup(handle);
  if(cbdata.get < 0)
    return(cbdata.get);
  if(get_ret != OBEX_STAT_SENT)
    return(-3);
  if(cbdata.get == 0)
    return(-2);
  return(0);
}

/* ---------------------------------------------------------------- */
/*
 * Send a squirt over a GET, Issue a get-interface command and
 * return the result, if possible...
 */
int
send_irda_obex_get(es_squirt *	squirt,		/* Optional squirt */
		   es_opt *	options,	/* For the squirt */
		   es_trans *	extra,		/* Transport data */
		   gi_obget *	request,	/* What to get */
		   gi_flags *	flags,		/* How to get it */
		   int		opt_ias)
{
  obex_t *		handle = NULL;	/* Obex connection handle */
  obex_object_t *	object;		/* Obex object to send */
  gi_obgetsend		cbdata;		/* Data for our callback... */
  es_obex		command;	/* The optional PUT command */
  int			obex_flags;
  int			ret = 0;
  int			get_ret = -1;	/* Return of the get command */

  map_obex_struct_fields(squirt, extra, &command);

  /* Create a handle for this connection */
  if(flags->filter == OPT_NO_FILTER)
    obex_flags = 0;
  else
    obex_flags = OBEX_FL_FILTERHINT;
  handle = OBEX_Init(OBEX_TRANS_IRDA, obget_client_event, obex_flags);
  if(handle == (obex_t *) NULL)
    {
      fprintf(stderr, "OBEX_Init failed\n");
      return(-1);
    }

  /* Set as user data of the Obex handle */
  cbdata.handle = handle;
  cbdata.status = OBEX_STAT_CONT;
  cbdata.get = 0;
  cbdata.request = request;
  cbdata.flags = flags;
  cbdata.extra = extra;
  OBEX_SetUserData(handle, (void *) &cbdata);

  /* Increase throughput */
  OBEX_SetTransportMTU(handle, OBEX_IRDA_OPT_MTU, OBEX_IRDA_OPT_MTU);

  /* Find a peer, and connect (socket) to it */
  if(opt_ias == OPT_OBEX_DO_STDIAS)
    ret = IrOBEX_TransportConnect(handle, OBEX_STD_SERVICE_NAME);
  else
    ret = IrOBEX_TransportConnect(handle, OBEX_SERVICE_NAME);
  if(ret < 0)
    {
      OBEX_Cleanup(handle);
      return(ret);
    }

  /* Creating a sesssion is optional. It improves performance for long
   * transaction and allow to do get_interface, but increases latency */
  if(flags->obex.session != OPT_OBEX_NO_SESSION)
    {
      /* Start the Obex session (optional - may want to add count hdr) */
      object = OBEX_ObjectNew(handle, OBEX_CMD_CONNECT);
      ret = obget_send_object(&cbdata, object);
      if(ret != OBEX_STAT_SENT)
	{
	  OBEX_Cleanup(handle);
	  return(-1);
	}
    }

  /* Those headers are reserved for the Get-Interface options */
  options->obex.name = OPT_OBEX_NO_NAME;
  if(flags->obex.target == OPT_OBEX_TARGET_USER)
    options->model = OPT_UNDEFINED;
  /* If we put a Squirt in the body, we can't put some request data as well
   * so disable it */
  if((flags->squirt_get != OPT_NO_SQUIRT_GET) &&
     (request->datalen >= 0) && (request->data != NULL))
    {
      fprintf(stderr, "Can't process request->data with a Squirt - dropped\n");
      free(request->data);
      request->data = NULL;
#ifdef DEBUG_ALLOC
      fprintf(stderr, "%s: free(request->data)\n", __FUNCTION__);
#endif
    }
  /* No type has a special meaning in GET requests, it means name overflow,
   * so don't generate it in a squirt ! */
  if(options->obex.type == OPT_OBEX_TYPE_NONE)
    options->obex.type = OPT_UNDEFINED;
  /* And if you use options->obex.type which is not REFS, REFX or REFL
   * are definitely asking for troubles... */

  /* Build Obex GET object containing the request parameters */
  object = OBEX_ObjectNew(handle, OBEX_CMD_GET);
  /* Add the Get-Interface headers */
  create_obget_headers(handle, object, request, flags, extra);
  /* Add the squirt header+payload (if not disabled) */
  if(flags->squirt_get != OPT_NO_SQUIRT_GET)
    create_obex_headers(handle, object, &command, options);

  /* Process the "get" (send it) */
  get_ret = obget_send_object(&cbdata, object);

  /* Sesssion is optional. */
  if(flags->obex.session != OPT_OBEX_NO_SESSION)
    {
      /* Finish the Obex session */
      object = OBEX_ObjectNew(handle, OBEX_CMD_DISCONNECT);
      ret = obget_send_object(&cbdata, object);
      /* We don't really care if this one fails */
    }

  /* The end - The data will be in the request struct, if any */
  OBEX_Cleanup(handle);
  if(cbdata.get < 0)
    return(cbdata.get);
  if(get_ret != OBEX_STAT_SENT)
    return(-3);
  if(cbdata.get == 0)
    return(-2);
  return(0);
}

/************************** OBEX RECEIVER **************************/
/*
 * This is where we actually receive beacons and e-Squirt
 * over the IrObex transport (regular Obex over TinyTP).
 * Under Linux, we simply use the excellent Obex support ;-)
 */

/* ---------------------------------------------------------------- */
/*
 * Handle the reception of some data on the Obex server
 * Just create a separate Obex handle via OBEX_ServerAccept()
 *
 * Note on Obex user data : as opposed to the send, the user data is
 * valid only within each individual function. It is therefore *mandatory*
 * to set it prior to any call to OBEX_HandleInput().
 */
int
irda_obex_accept(es_listen *	desc,
		 es_rx_data *	rx_data)
{
  es_obexrecv		cbdata;		/* Data for our callback... */
  int			ret;

  /* Set our stuff as user data of the Obex handle */
  cbdata.handle = desc->handle;
  cbdata.status = OBEX_STAT_CONT;
  OBEX_SetUserData(desc->handle, (void *) &cbdata);

  /* Get answer from Obex lib */
  while(cbdata.status == OBEX_STAT_CONT)
    {
      ret = OBEX_HandleInput(desc->handle, 1);
      if(ret < 0)
	return(ret);
    }
  /* Check if we got what we want */
  if(cbdata.status != OBEX_STAT_ACCEPT)
    {
      fprintf(stderr, "Invalid status 0x%X\n", cbdata.status);
      return(-1);
    }

  /* Reset handle data back to default */
  OBEX_SetUserData(desc->handle, (void *) NULL);

  /* Get a new handle for the new connection */
  desc->rx_handle = OBEX_ServerAccept(desc->handle, obex_recv_event, NULL);

  return(0);
}

/* ---------------------------------------------------------------- */
/*
 * Post processing handler for Obex socket...
 * Receive and parse the data appropriately...
 */
int
irda_obex_recv(es_listen *	desc,
	       es_rx_data *	rx_data)
{
  es_squirt *		content = rx_data->content;
  es_opt *		options = rx_data->options;
  es_trans *		extra   = rx_data->extra;
  es_obexrecv		cbdata;		/* Data for our callback... */
  int			ret;

  /* Set our stuff as user data of the Obex handle */
  cbdata.handle = desc->rx_handle;
  cbdata.status = OBEX_STAT_CONT;
  cbdata.put = -1;
  cbdata.get = -1;
  cbdata.session = OPT_OBEX_NO_SESSION;
  map_obex_struct_fields(content, extra, &(cbdata.data));
  cbdata.options = options;
  cbdata.userdata = rx_data;
  OBEX_SetUserData(desc->rx_handle, (void *) &cbdata);

  /* Set some options */
  options->transport = OPT_IRDA;
  options->protocol = OPT_OBEX;
  options->model = OPT_UNDEFINED;
  options->obex.target = OPT_UNDEFINED;
  options->obex.ias = desc->opt;
  options->obex.name = OPT_OBEX_NO_NAME;
  options->obex.type = OPT_OBEX_TYPE_NONE;
  options->obex.session = OPT_OBEX_NO_SESSION;

  /* Get answer from Obex lib */
  while(cbdata.status == OBEX_STAT_CONT)
    {
      ret = OBEX_HandleInput(desc->rx_handle, 1);
      if(ret < 0)
	break;
    }

  /* Note : at this point, both Obex and XML should be properly
   * decapsulated. See process_put_request() for details...
   * In most cases, we even have called the user callback... */
  return(cbdata.put);	/* What finish_receive() expects */
}

/* ---------------------------------------------------------------- */
/*
 * Destroy the Obex receiver. Cleanup its handle.
 */
int
irda_obex_end_recv(es_listen *	desc)
{
  /* Close the handle */
  if(desc->rx_handle != NULL)
    OBEX_Cleanup(desc->rx_handle);
  desc->rx_handle = 0;
  return(0);
}

/* ---------------------------------------------------------------- */
/*
 * Destroy a Obex listener. Cleanup the handle and free pieces.
 */
int
irda_obex_destroy(es_listen *	desc)
{
  /* Close the handles */
  if(desc->handle != NULL)
    OBEX_Cleanup(desc->handle);
  if(desc->rx_handle != NULL)
    OBEX_Cleanup(desc->rx_handle);
  /* Free pieces */
  desc->next = (es_listen *) NULL;
  free(desc);
  return(0);
}

/* ---------------------------------------------------------------- */
/*
 * Add a receiver description for IrDA socket to the list
 */
es_listen *
add_irda_obex(es_listen *	list,
	      es_opt *		options,
	      int		opt_ias)
{
  es_listen *		desc;
  obex_t *		handle = NULL;	/* Obex connection handle */
  int			sock;		/* Socket itself */
  int			ret;

  /* Create a handle for this connection */
  handle = OBEX_Init(OBEX_TRANS_IRDA, obex_server_event, OBEX_FL_KEEPSERVER);
  if(handle == (obex_t *) NULL)
    {
      fprintf(stderr, "OBEX_Init failed\n");
      return(NULL);
    }

  /* Increase throughput */
  OBEX_SetTransportMTU(handle, OBEX_IRDA_OPT_MTU, OBEX_IRDA_OPT_MTU);

  /* Start the Obex server */
  if(opt_ias == OPT_OBEX_DO_STDIAS)
    ret = IrOBEX_ServerRegister(handle, OBEX_STD_SERVICE_NAME);
  else
    ret = IrOBEX_ServerRegister(handle, OBEX_SERVICE_NAME);
  if(ret < 0)
    {
      fprintf(stderr, "OBEX : Unable to register the socket (no IrDA stack ?)\n");
      return(NULL);
    }

  /* Need to get the socket here !!! */
  sock = OBEX_GetFD(handle);

  /* Allocate some memory for the receiver description */
  desc = (es_listen *) malloc(sizeof(es_listen));
  if(desc == NULL)
    {
      fprintf(stderr, "Malloc failed\n");
      return(NULL);
    }

  /* Fill the description */
  desc->sock = sock;
  desc->rx_sock = -1;
  desc->opt = opt_ias;
  desc->ultra = 0;
  desc->handle = (void *) handle;
  desc->rx_handle = (void *) NULL;
  desc->pre_handler = &irda_obex_accept;
  desc->post_handler = &irda_obex_recv;
  desc->end_recv = &irda_obex_end_recv;
  desc->destroy = &irda_obex_destroy;

  /* Link to the list and goodbye */
  desc->next = list;
  return(desc);
}
