/*
 *			    Squirt Obex PUT common
 *			    ----------------------
 *
 *		Jean II '00
 *
 * Handle all the common Obex Put code for the various e-squirt applications.
 *
 * This file is released under the GPL license.
 * Note : this file is inspired by code written by :
 *	o Dag Brattli <dagb@fast.no>
 *	o Pontus Fuchs <pontus.fuchs@tactel.se> 
 * Thanks to them ;-)
 */

/*
Copyright (C) 1995-2001 Hewlett-Packard Company 

This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation; either version 2 of the License, or (at your option) any later version. 

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details. 

You should have received a copy of the GNU General Public License along with this program; if not, write to the Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. 
*/

#include "squirt_obex.p.h"

/*************************** OBEX SENDER ***************************/
/*
 * Sending an object over Obex...
 * Call back for the Obex library and all that jazz...
 */

/* ---------------------------------------------------------------- */
/*
 * Create a Obex Beacon or E-Squirt headers
 *
 * IMPORTANT : this is just a test sample, and doesn't represent how
 * a beacon or e-Squirt should be, because Obex is flexible.
 * For example, we might remove the type header (optional) and the headers
 * may be swapped around.
 * The string is either a plain URL or a XML description.
 */
int
create_obex_headers(obex_t *		handle,
		    obex_object_t *	object,
		    es_obex *		content,
		    es_opt *		options)
{
  obex_headerdata_t	hdd;		/* Header data */
  int			len;

  /* Check if beacon or e-Squirt */
  hdd.bs = get_es_obex_target(options, &len);
  if(hdd.bs != NULL)
    {
      OBEX_ObjectAddHeader(handle, object, OBEX_HDR_TARGET,
			   hdd, len, 0);
    }

  /* Add the type header (optional) note : need '\0' */
  hdd.bs = get_es_obex_type(options, content->usertype);
  if(hdd.bs != NULL)
    {
      OBEX_ObjectAddHeader(handle, object, OBEX_HDR_TYPE,
			   hdd, strlen(hdd.bs) + 1, 0);
    }

  /* Add the name header (optional) 
   * note : char[] -> unicode, and need '\0' */
  if(options->obex.name == OPT_OBEX_HAS_NAME)
    {
      uint8_t	ucname[ES_MAX_FILENAME * 2];
      int	ucname_len = strlen(content->filename)*2 + 2;

      ucname_len = OBEX_CharToUnicode(ucname, content->filename, ucname_len);
      hdd.bs = ucname;
      OBEX_ObjectAddHeader(handle, object, OBEX_HDR_NAME, hdd, ucname_len, 0);
    }
  /* Add the name header (optional)
   * note : pure unicode, and need '\0\0' */
  if(options->obex.name == OPT_OBEX_UNICODE_NAME)
    {
      hdd.bs = content->filename;
      OBEX_ObjectAddHeader(handle, object, OBEX_HDR_NAME,
			   hdd, unicode_strlen(hdd.bs)*2 + 2, 0);
    }

  /* Add the payload (the "end of body" flag is automatic...) */
  hdd.bs = content->payload;
  OBEX_ObjectAddHeader(handle, object, OBEX_HDR_BODY,
		       hdd, content->size, 0);

  return(0);
}

/* ---------------------------------------------------------------- */
/*
 * Even CallBack from the Obex library
 * Only used on the sender side...
 */
void
obex_client_event(obex_t *		handle,
		  obex_object_t *	object,
		  int			mode,
		  int			event,
		  int			obex_cmd,
		  int			obex_rsp)
{
  es_obexsend *		cbdata;

  cbdata = OBEX_GetUserData(handle);
  if(cbdata == NULL)
    return;
  cbdata->status = OBEX_STAT_ERROR;	/* Default : error... */

  switch(event)
    {
    case OBEX_EV_PROGRESS:
      /* This one is usually never called */
      cbdata->status = OBEX_STAT_CONT;
      break;
    case OBEX_EV_REQDONE:
      /* Comes when a command has finished. */
      if(obex_rsp == OBEX_RSP_SUCCESS)
	{
	  cbdata->status = OBEX_STAT_SENT;

	  /* Check if it was a PUT */
	  if(obex_cmd == OBEX_CMD_PUT)
	    {
	      /* Put has been done */
	      cbdata->put = 1;

	      /* If the user want the returned headers, try to extract them */
	      if(cbdata->reply != (gi_obget *) NULL)
		{
		  parse_obget_header(handle, object,
				     cbdata->reply, cbdata->flags,
				     cbdata->extra, cbdata->options);
		  /* A valid reply (for us) must contain a name */
		  cbdata->has_replied = (cbdata->flags->obex.name >=
					 OPT_OBEX_HAS_NAME);
		}
	    }
	}
      else
	{
	  fprintf(stderr, "Obex command not accepted\n");
	  /* Check if it was a PUT */
	  if(obex_cmd == OBEX_CMD_PUT)
	    cbdata->put = -(obex_rsp);
	}
      break;
    case OBEX_EV_LINKERR:
      fprintf(stderr, "Link broken\n");
      break;
    default:
      fprintf(stderr, "Unknown event 0x%X\n", event);
      break;
    }
}

/* ---------------------------------------------------------------- */
/*
 * Perform the Obex request and process the answer...
 */
int
obex_send_object(es_obexsend *		cbdata,
		 obex_object_t *	object)
{
  int ret;

  /* Reset our flag */
  cbdata->status = OBEX_STAT_CONT;

  /* Do the request */
  OBEX_Request(cbdata->handle, object);

  /* Get answer from Obex lib */
  while(cbdata->status == OBEX_STAT_CONT)
    {
      ret = OBEX_HandleInput(cbdata->handle, 1);
      if(ret < 0)
	return(ret);
    }
  return(cbdata->status);
}


/*************************** OBEX SERVER ***************************/
/*
 * Wait for incomming connections over the Obex server
 */

/* ---------------------------------------------------------------- */
/*
 * Even CallBack from the Obex library
 * Only used on the server side to wait for an accept...
 */
void
obex_server_event(obex_t *		handle,
		  obex_object_t *	object,
		  int			mode,
		  int			event,
		  int			obex_cmd,
		  int			obex_rsp)
{
  es_obexrecv *		cbdata;

  cbdata = OBEX_GetUserData(handle);
  if(cbdata == NULL)
    return;
  cbdata->status = OBEX_STAT_ERROR;	/* Default : error... */

  switch(event)
    {
    case OBEX_EV_PROGRESS:
      /* This one is usually never called */
      cbdata->status = OBEX_STAT_CONT;
      break;
    case OBEX_EV_LINKERR:
      fprintf(stderr, "Link broken\n");
      break;
    case OBEX_EV_ACCEPTHINT:
      /* This is the only valid event we should get */
      cbdata->status = OBEX_STAT_ACCEPT;
      break;

    default:
      fprintf(stderr, "Unknown event 0x%X\n", event);
      break;
    }
}

/************************** OBEX RECEIVER **************************/
/*
 * After the server has accepted the incomming connection, this part
 * handle the reception
 */

/* ---------------------------------------------------------------- */
/*
 * Even CallBack from the Obex library
 * Handle all commands sent by other side, like PUT, GET and co...
 */
int
parse_obex_object(obex_t *		handle,
		  obex_object_t *	object,
		  es_obex *		data,
		  es_opt *		options,
		  es_opt *		listen_options)
{
  obex_headerdata_t	hv;
  uint8_t		hi;
  uint32_t		hlen;
  int			done = 0;
  int			ret;

  /* Process all headers */
  while(OBEX_ObjectGetNextHeader(handle, object, &hi, &hv, &hlen))
    {
      switch(hi)
	{
	case OBEX_HDR_TARGET:
	  /* Check if it is one of the various targets we know about */
	  match_es_obex_target(hv.bs, hlen, options);
	  break;
	case OBEX_HDR_TYPE:
	  /* Same for type */
	  match_es_obex_type(hv.bs, hlen, options, data->usertype);
	  break;
	case OBEX_HDR_NAME:
	  ret = -2;
	  /* Check what we do with regard to Unicode */
	  if(listen_options->obex.name == OPT_OBEX_UNICODE_NAME)
	    {
	      /* No Unicode translation */
	      if(hlen < (ES_MAX_FILENAME - 1))
		ret = unicode_strcpy(data->filename, hv.bs, hlen/2);
	      options->obex.name = OPT_OBEX_UNICODE_NAME;
	    }
	  else
	    {
	      /* Translate Unicode */
	      if((hlen/2) < ES_MAX_FILENAME)
		ret = unicode_to_char_strcpy(data->filename, hv.bs, hlen/2);
	      options->obex.name = OPT_OBEX_HAS_NAME;
	    }
	  /* Error */
	  if(ret < 0)
	    {
	      fprintf(stderr, "Name too big (%d)...\n", hlen/2);
	      options->obex.name = OPT_UNDEFINED;	/* == error */
	    }
	  break;
	case OBEX_HDR_HTTP:
	  /* If we are called from process_get_request() - ignore */
	  break;
	case OBEX_HDR_BODY:
	  /* Do we have a buffer ? */
	  if(data->payload == NULL)
	    {
	      /* Allocate the right size */
	      data->payload = malloc(hlen + 1);
	      if(data->payload != NULL)
		{
		  memcpy(data->payload, hv.bs, hlen);
		  data->payload[hlen] = '\0';	/* Terminate the string */
		  data->size = hlen;
		  done = 1;
#ifdef DEBUG_ALLOC
		  fprintf(stderr, "%s : malloc(%d)\n", __FUNCTION__, hlen + 1);
#endif
		}
	      else
		fprintf(stderr, "Obex Payload malloc failed...\n");
	    }
	  else
	    {
	      /* Check size */
	      if(hlen < (data->size - 1))
		{
		  memcpy(data->payload, hv.bs, hlen);
		  data->payload[hlen] = '\0';	/* Terminate the string */
		  data->size = hlen;
		  done = 1;
		}
	      else
		fprintf(stderr, "Obex Payload too big (%d)...\n", hlen);
	    }
	  break;
	default:
	  fprintf(stderr, "Skipped header %02x\n", hi);
	}
    }

  return(done);
}

/* ---------------------------------------------------------------- */
/*
 * Set the headers of a PUT reply
 * Important note : the PUT reply *must* fit in one Obex packet. That is
 * the only Obex packet with such a constraint.
 */
int
create_obex_reply(obex_t *		handle,
		  obex_object_t *	object,
		  gi_obget *		request,
		  gi_flags *		flags,
		  es_trans *		extra)
{
  obex_headerdata_t	hdd;		/* Header data */
  char			ucname[ES_MAX_FILENAME * 2];	/* Name in unicode */
  int			hdr_len;

  /* Check if there is a name - ASCII */
  if(flags->obex.name == OPT_OBEX_HAS_NAME)
    {
      /* Convert filename in unicode */
      hdr_len = OBEX_CharToUnicode(ucname, extra->filename,
				      sizeof(ucname));

      /* Add it to the header */
      hdd.bs = ucname;
      OBEX_ObjectAddHeader(handle, object, OBEX_HDR_NAME, hdd,
			   hdr_len, OBEX_FL_FIT_ONE_PACKET);
    }
  /* Check if there is a name - Unicode */
  if(flags->obex.name == OPT_OBEX_UNICODE_NAME)
    {
      hdd.bs = extra->filename;
      OBEX_ObjectAddHeader(handle, object, OBEX_HDR_NAME,
			   hdd, unicode_strlen(hdd.bs)*2 + 2,
			   OBEX_FL_FIT_ONE_PACKET);
    }

  /* Check if there is an HTTP header (may not fit) */
  hdd.bs = build_gi_obex_http(flags, extra);
  if(hdd.bs != NULL)
    {
      OBEX_ObjectAddHeader(handle, object, OBEX_HDR_HTTP, hdd,
			   strlen(hdd.bs), OBEX_FL_FIT_ONE_PACKET);

      /* Cleanup before returning to the main server */
      if(extra->http_cleanup == OPT_DATA_DO_FREE)
	{
	  free(extra->http_hdr);
#ifdef DEBUG_ALLOC
	  fprintf(stderr, "%s: free(extra->http_hdr)\n", __FUNCTION__);
#endif
	}
      extra->http_hdr = NULL;
    }

  /* The user should not pass us any data, but... */
  if((request->data != NULL) &&
     (request->cleanup == OPT_DATA_DO_FREE))
    {
      free(request->data);
#ifdef DEBUG_ALLOC
      fprintf(stderr, "%s: free(reply.data)\n", __FUNCTION__);
#endif
    }
  request->data = NULL;

  return(0);
}

/* ---------------------------------------------------------------- */
/*
 * When we receive an incomming PUT, obex_recv_event() will call this
 * function. We decapsulate the squirt and tell if it's ok.
 * We will even try to call the user calback so that we can return
 * proper headers in our reply...
 * Note on the return value :
 *	-1 == error
 *	 0 == ok, callback not called (so the user has not got the data yet)
 *	 1 == ok, callback called (the user has already a copy of the data)
 */
int
process_put_request(obex_t *		handle,
		    obex_object_t *	object,
		    es_obex *		data,
		    es_opt *		options,
		    es_rx_data *	userdata)
{
  int			validput;
  int			ret = -1;

  /* Decapsulate e-Squirt data */
  validput = parse_obex_object(handle, object, data, options,
			       userdata->listen_options);

  /* Do we have anything valid ? */
  if(validput)
    {
      /* Success !!! Set the Obex response code */
      OBEX_ObjectSetRsp(object, OBEX_RSP_CONTINUE, OBEX_RSP_SUCCESS);

      /* Hook the receive buffer */
      if(userdata->content->data == NULL)
	{
	  userdata->content->data = data->payload;
	  userdata->content->cleanup = OPT_DATA_DO_FREE;
	}
      userdata->content->datalen = data->size;

      /* Call the user callback if possible */
      if(userdata->callback != (squirt_handler) NULL)
	{
	  gi_obget	reply;		/* Headers we return */
	  gi_flags	flags;		/* Flags to return */

	  /* Quick init */
	  memset(&reply, 0, sizeof(reply));
	  memset(&flags, 0, sizeof(flags));
	  /* Set our flags */
	  flags.protocol = OPT_OBEX;
	  flags.transport = data->transport;
	  flags.obex.session = options->obex.session;
	  flags.obex.ias = options->obex.ias;

	  /* Process XML, verify model and Call user */
	  ret = call_rcv_callback(userdata, &reply, &flags);
	  /* Check if callback was successful... */
	  if(ret >= 0)
	    validput = 2;	/* Callback called */
	  else
	    {
	      /* Failure (or should we use OBEX_RSP_NO_CONTENT ?) */
	      int	obex_err = OBEX_RSP_BAD_REQUEST;
	      if(ret < -10)
		obex_err = -ret;
	      OBEX_ObjectSetRsp(object, obex_err, obex_err);
	      validput = ret;	/* Failure */
	    }
	  /* If callback return 1 or 2, we have some headers */
	  if(ret > 0)
	    {
	      /* Create replay, cleanup user data */
	      create_obex_reply(handle, object, &reply, &flags,
				userdata->extra);
	    }

	  /* In parse_obex_object(), we may allocate a buffer (data->payload),
	   * which may be freed in call_rcv_callback(), or the callback
	   * may change it. So, get back the modified values. */
	  data->payload = userdata->content->data;
	  data->size = userdata->content->datalen;
	}
    }
  else
    {
      /* Failure (or should we use OBEX_RSP_NO_CONTENT ?) */
      OBEX_ObjectSetRsp(object, OBEX_RSP_BAD_REQUEST, OBEX_RSP_BAD_REQUEST);
    }

  return(validput - 1);
}


/* ---------------------------------------------------------------- */
/*
 * Even CallBack from the Obex library
 * Handle all commands sent by other side, like PUT, GET and co...
 */
void
obex_recv_event(obex_t *		handle,
		obex_object_t *		object,
		int			mode,
		int			event,
		int			obex_cmd,
		int			obex_rsp)
{
  es_obexrecv *		cbdata;
  es_obex *		data;
  es_opt *		options;
  es_rx_data *		userdata;

  cbdata = OBEX_GetUserData(handle);
  if(cbdata == NULL)
    return;
  cbdata->status = OBEX_STAT_CONT;	/* Default : continue... */
  data = &(cbdata->data);
  options = cbdata->options;
  userdata = cbdata->userdata;

  switch(event)
    {
    case OBEX_EV_PROGRESS:
      break;
    case OBEX_EV_LINKERR:
      cbdata->status = OBEX_STAT_ERROR;
      fprintf(stderr, "Link broken\n");
      break;

    case OBEX_EV_REQHINT:
      /* Comes BEFORE the lib parses anything, just at the point when
       * it has received the Obex Connect packet */
      switch(obex_cmd)
	{
	case OBEX_CMD_SETPATH:
	  fprintf(stderr, "Bug - OBEX_CMD_SETPATH\n");
	case OBEX_CMD_PUT:
	case OBEX_CMD_CONNECT:
	case OBEX_CMD_DISCONNECT:
	  /* Just accept the incomming Obex connection */
	  OBEX_ObjectSetRsp(object, OBEX_RSP_CONTINUE, OBEX_RSP_SUCCESS);
	  break;
	case OBEX_CMD_GET:
	  /* Check if a GET callback is activated */
	  if(check_get_request(handle, object))
	    /* Just accept the incomming Obex connection */
	    OBEX_ObjectSetRsp(object, OBEX_RSP_CONTINUE, OBEX_RSP_SUCCESS);
	  else
	    OBEX_ObjectSetRsp(object, OBEX_RSP_NOT_IMPLEMENTED,
			      OBEX_RSP_NOT_IMPLEMENTED);
	  break;
	default:
	  /* Refuse the Obex connection */
	  fprintf(stderr, "Skipping unsupported command:%02x\n", obex_cmd);
	  OBEX_ObjectSetRsp(object, OBEX_RSP_NOT_IMPLEMENTED,
			    OBEX_RSP_NOT_IMPLEMENTED);
	  break;
	}
      break;

    case OBEX_EV_REQ:
      /* Comes when a server-request has been received.
       * We do the bulk of the processig here... */
      switch(obex_cmd)
	{
	case OBEX_CMD_CONNECT:
	  /* Set our private flag */
	  cbdata->session = OPT_OBEX_DO_SESSION;
	  /* Set flag for the user */
	  options->obex.session = OPT_OBEX_DO_SESSION;
	  break;
	case OBEX_CMD_DISCONNECT:
	  break;
	case OBEX_CMD_PUT:
	  /* If not data received yet or callback already called,
	   * decapsulate e-Squirt data */
	  if(cbdata->put != 0)
	    cbdata->put = process_put_request(handle, object,
					      data, options, userdata);
	  break;
	case OBEX_CMD_GET:
	  /* If not data received yet or callback already called,
	   * Reply to the GET request */
	  if(cbdata->put != 0)
	    cbdata->put = process_get_request(handle, object,
					      data, options, userdata);
	  break;
	case OBEX_CMD_SETPATH:
	  fprintf(stderr, "Bug - OBEX_CMD_SETPATH\n");
	  break;
	default:
	  fprintf(stderr, "Received unsupported command:%02x\n", obex_cmd);
	  break;
	}
      break;

    case OBEX_EV_REQDONE:
      /* Comes when a command has finished. */
      if(obex_cmd == OBEX_CMD_DISCONNECT)
	cbdata->status = OBEX_STAT_RECV;
      /* Handle no session Obex gracefully
       * Note : use our private flag, as the user may have changed his own */
      if((cbdata->session != OPT_OBEX_DO_SESSION) &&
	 (obex_cmd != OBEX_CMD_CONNECT))
	cbdata->status = OBEX_STAT_RECV;
      break;

    default:
      fprintf(stderr, "Unknown event 0x%X\n", event);
      break;
    }
}
