/*
 *			    e-Squirt to HTTP proxy
 *			    ----------------------
 *
 *		Jean II '99, '00, '01 and '02
 *
 * Listen to e-Squirt requests over IrDA and convert them to HTTP requests.
 * 
 * This file is released under the GPL license.
 */

/*
Copyright (C) 1995-2001 Hewlett-Packard Company 

This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation; either version 2 of the License, or (at your option) any later version. 

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details. 

You should have received a copy of the GNU General Public License along with this program; if not, write to the Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. 
*/

#include <sys/types.h>
#include <sys/wait.h>

#include <unistd.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <signal.h>

#include <squirt.h>
#include <http.h>

/**************************** CONSTANTS ****************************/

const char	esurl[] = "http://127.0.0.1/esquirt";
const char	opurl[] = "http://127.0.0.1/obexput/";

/*********************** HTTP HEADER MANGLING ***********************/
/*
 */

/* Each header 'type' *must* be terminated by a ':', otherwise you may
 * match the wrong header */
static const char *	http_type[] = { "Set-Cookie:", "Cookie:" };
static const int	http_type_num = sizeof(http_type)/sizeof(http_type[0]);

/* ---------------------------------------------------------------- */
/*
 * Strip unwanted headers
 *
 * We do the operation 'in-place", so the original header is lost...
 */
static int
http_strip_headers(char *	header)
{
  char *	roff = header;		/* Read offset */
  char *	woff = header;		/* Write offset */
  char *	end = header + strlen(header);
  char *	eol;
  int		len;
  int		i;

  do      /* For each line of header */
    {
      /* Find if a tag matches */
      for(i = 0; i < http_type_num; i++)
	if(!strncasecmp(roff, http_type[i], strlen(http_type[i])) )
	  break;

      /* Find end of line */
      eol = strchr(roff, '\n');

      /* If we got a match */
      if(i < http_type_num)
	{
	  /* Will not fail, libhttp always keep '\n' */
	  if(eol != NULL)
	    {
	      /* Keep "\r\n", copy */
	      len = eol - roff + 1;
	      memcpy(woff, roff, len);
	      woff += len;
	    }
        }

      /* Skip this line */
      roff = eol;
    }       /* Go to start of next line, if any */
  while((roff != NULL) && (++roff < end));

  /* Terminate string */
  *woff = '\0';

  return(woff != header);
}

/************************** HTML MANGLING **************************/
/*
 * Processing of HTML pages. We mostly want to fix the embedded URLs
 * so that the other end does the right thing...
 */

/* ---------------------------------------------------------------- */
/*
 * Get path out of a name
 *
 * Just truncate everything after the last '/'
 * Also return the hostname
 */
static inline void
get_base_path(char *	basename,
	      char *	hostname)
{
  char *	p;
  int		l = strlen(basename);
  char *	e;

  /* Check if name contains the host name */ 
  p = strstr(basename, "://");
  if(p)
    {
      /* Find end of hostname */
      p += 3;
      p = strchr(p, '/');
      if(p)
	l = p - basename;
      else
	p = basename + l;
      memcpy(hostname, basename, l);
      hostname[l] = '\0';
    }
  else
    {
      hostname[0] = '\0';
      p = basename;
    }

  /* Find end of the path */
  e = strrchr(p, '/');
  if(!e)
    e = p;

  /* Remove last component */
  *e = '\0';
}

/* ---------------------------------------------------------------- */
/*
 * Mangle a URL
 *
 * Add the "obex:/" prefix to each URL so that it goes back to us.
 * Also change relative URLs to absolute URLs.
 *
 * The source is a non-NULL terminated string.
 * Destination buffer plus its size.
 * Return lenght of new page.
 */
static char *
mangle_url(char *	d,		/* Destination */
	   char *	s,		/* Source */
	   char *	path,		/* Base URL (for relative URLs) */
	   char *	host)		/* Host name (for relative URLs) */
{
  /* Check if the URL is not relative to the page itself */
  if(*s != '#')
    {
      /* Absolute URLs */
      if(!strncasecmp(s, "http:", 5))
	{
	  /* Insert "obex:/" prefix so that we can trap it */
	  memcpy(d, "obex:/", 6);
	  d += 6;
	}
      else
	{
	  char *q = strchr(s, '"');
	  char *c = strchr(s, ':');

	  if((c != NULL) && (c < q))
	    {
	      /* Other protocol, such as "ftp:", "mailto:" :
	       * We don't support it, so don't touch... */
	      ;
	    }
	  else
	    {
	      if(*s == '/')
		{
		  /* URL relative to the current server, make it absolute. */
		  memcpy(d, "obex:/", 6);
		  d += 6;
		  strcpy(d, host);
		  d += strlen(host);
		}
	      else
		{
		  /* URL relative to the current path, make it absolute. */
		  memcpy(d, "obex:/", 6);
		  d += 6;
		  strcpy(d, path);
		  d += strlen(path);
		  *d = '/';
		  d++;
		}
	    }
	}
      /* Finito ! */
    }
  return(d);
}

/* ---------------------------------------------------------------- */
/*
 * Mangle an HTML page
 *
 * Check every URL and add the "obex:/" prefix so that it goes back to us.
 * Also change relative URLs to absolute URLs.
 * You can call this function with binary data, in 99.999% of cases it
 * won't touch it.
 *
 * The source is a non-NULL terminated string.
 * Destination buffer plus its size.
 * Return lenght of new page.
 */
static int
mangle_html(char *	dest,		/* Destination */
	    int		dlen,
	    char *	src,		/* Source */
	    int		slen,
	    char *	path,		/* Base URL (for relative URLs) */
	   char *	host)		/* Host name (for relative URLs) */
{
  char *	s = src;
  char *	d = dest;
  char *	e = dest + dlen;
  char		last;

  /* Save last char of the source, replace with a NULL */
  last = src[slen - 1];
  src[slen - 1] = '\0';

  /* Look every character of the page */
  while((*s != '\0') && (d < e))
    {
      /* Is it a HTML tag ? */
      if(*s != '<')
	{
	  /* Just copy */
	  *d = *s;
	  d++;
	  s++;
	}
      else
	{
	  /* Copy */
	  *d = *s;
	  d++;
	  s++;

	  /* -------------- Check for an anchor -------------- */
	  if(((*s == 'a') || (*s == 'A')) && (isspace(s[1])))
	    {
	      /* Skip blanks (we should use isblank()) */
	      do
		{
		  *d = *s;
		  d++;
		  s++;
		}
	      while(isspace(*s));

	      /* Check for href */
	      if(!strncasecmp(s, "href", 4))
		{
		  /* Move to the URL */
		  do
		    {
		      *d = *s;
		      d++;
		      s++;
		    }
		  while(*s && (*s != '>') && (*(s - 1) != '"'));

		  if(*(s - 1) == '"')
		    d = mangle_url(d, s, path, host);
		}
	    }

	  /* -------------- Check for IMG and SCRIPT tags -------------- */
	  if((!strncasecmp(s, "img", 3)) ||
	     (!strncasecmp(s, "script", 6)))
	    {
	      /* Go to source attribute */
	      do
		{
		  *d = *s;
		  d++;
		  s++;
		}
	      while((*s != '>') && (strncasecmp(s, "src", 3)));

	      /* If we found a src attribute */
	      if(*s != '>')
		{
		  /* Move to the URL */
		  do
		    {
		      *d = *s;
		      d++;
		      s++;
		    }
		  while(*s && (*s != '>') && (*(s - 1) != '"'));

		  if(*(s - 1) == '"')
		    d = mangle_url(d, s, path, host);
		}
	    }

	  /* -------------- Check for FORM tags -------------- */
	  if(!strncasecmp(s, "form", 4))
	    {
	      /* Go to source attribute */
	      do
		{
		  *d = *s;
		  d++;
		  s++;
		}
	      while((*s != '>') && (strncasecmp(s, "action", 6)));

	      /* If we found a src attribute */
	      if(*s != '>')
		{
		  /* Move to the URL */
		  do
		    {
		      *d = *s;
		      d++;
		      s++;
		    }
		  while(*s && (*s != '>') && (*(s - 1) != '"'));

		  if(*(s - 1) == '"')
		    d = mangle_url(d, s, path, host);
		}
	    }
	}
    }

  /* Check errors */
  if(*s == '\0')
    {
      int		l;

      /* Take care of binary data */
      l = (slen - 1) - (s - src);
      memcpy(d, s, l);			/* Or cable */
      d += l;
      /* And of last char */
      *d = last;
      d++;
      *d = '\0';
      return(d - dest);
    }
  else
    return(0);
}



/************************* SIGNAL HANDLING *************************/

/* ---------------------------------------------------------------- */
/*
 * Catch signal genarated by our childs upon termination...
 */
void
rcv_signal(int	signum)
{
  int	status;
  int	err;

  /* Which signal ? */
  switch(signum)
    {
    case SIGCHLD:
      /* Wait for all the childs to die */
      do
	{
	  err = waitpid(-1, &status, WNOHANG);
#ifdef DEBUG_CHILD
	  if(err > 0)
	    printf("Child died, pid %d...\n", err);
#endif
	}
      while(err > 0);

      /* Need to reset the handler */
      signal(SIGCHLD, &rcv_signal);
      break;
    default:
    }
}

/* ---------------------------------------------------------------- */
/*
 * Set the signal handlers as we would like them...
 */
void
set_signal_handlers(void)
{
  signal(SIGCHLD, &rcv_signal);
}

/* ---------------------------------------------------------------- */
/*
 * Handler for timeout or signals in the select
 */
int
server_select(int		code,
	      void *		cbdata)		/* CallBack data */
{
#ifdef DEBUG_CHILD
  printf("Select did return %d\n", code);
#endif
  /* Here we could check if we need to exit the receiver main loop */
  return(0);		/* Continue */
}

/***************************** RECEIVER *****************************/

/* ---------------------------------------------------------------- */
/*
 * Receive a e-Squirt of Beacon. Forking.
 *
 * This is a "post-forking" model. All the people implementing HTTP servers
 * will cringe. Yeah, the rate of request we can expect from IrDA and
 * BlueTooth doesn't justify to have a high performance design, we
 * just want to avoid blocking on slow clients.
 */
int
server_accept(es_post *		post_data,	/* Opaque */
	      es_rx_data *	rx_data,	/* Content, option and co */
	      void *		cbdata)		/* CallBack data */
{
  int		ret = -1;
  int		child_pid = 0;

  /* Fork a child */
  child_pid = fork();

  /* Setup thing for post fork */
  post_fork_squirt(post_data, rx_data, child_pid);

  /* Perform post processing */
  ret = post_process_squirt(post_data, rx_data);

  /* Are we the child ? */
  if(child_pid == 0)
    {
      /* Do never return to the server in the child !!! */
      exit(0);
    }

  printf("Accept done, forked pid %d...\n", child_pid);

  /* Not the child : Return back to the server, 0 -> continue */
  return(0);
}

/* ---------------------------------------------------------------- */
/*
 * Receive a e-Squirt of Beacon. Non forking.
 */
int
receive_esquirt(es_squirt *	fields,		/* Rx payload */
		es_opt *	options,	/* Rx options */
		es_trans *	extra,		/* Rx transport data */
		gi_obget *	ret_hdr,	/* Not used */
		gi_flags *	ret_flags)	/* Not used */
{
  int		ret = 0;	/* Reply Header not set */
  char *	url;		/* URL to query */
  int		urllen;		/* Length of URL */
  char		hdr[256];	/* POST header to send */
  HTTP_Extra	http;		/* HTTP header + POST data */
  HTTP_Response reply;		/* HTTP reply */
  const char	hdrstart[] = "Content-Type: ";
  int		startlen;

  // Binary == yuck...
  debug_squirt(fields, options, extra);

  memset(&http, '\0', sizeof(http));

  /* Deal with the type */
  if((options->obex.type != OPT_UNDEFINED) &&
     (options->obex.type != OPT_OBEX_TYPE_NONE))
    {
      /* Obex type is alway written in extra->type */
      startlen = strlen(hdrstart);
      memcpy(hdr, hdrstart, startlen);
      strcpy(hdr + startlen, extra->type);
      http.Headers = hdr;
    }
  else
    {
      /* Empty header */
      hdr[0] = '\0';
    }

  /* Check if it's a standard OBEX PUT and not a reference */
  if((options->protocol == OPT_OBEX) &&
     //(options->obex.name == OPT_OBEX_HAS_NAME) &&
     !(options->obex.type & (OPT_OBEX_TYPE_REFS |
			     OPT_OBEX_TYPE_REFX |
			     OPT_OBEX_TYPE_REFL)) )
    {
      /* Create Obex PUT request (as a POST) */
      startlen = strlen(opurl);

      /* Check if there is a name */
      if(options->obex.name == OPT_OBEX_HAS_NAME)
	urllen = startlen + strlen(extra->filename);
      else
	urllen = startlen;

      url = malloc(urllen + 1);
      memcpy(url, opurl, startlen + 1);

      /* Add filename as needed, after the '/' */
      if(options->obex.name == OPT_OBEX_HAS_NAME)
	memcpy(url + startlen, extra->filename,
	       strlen(extra->filename) + 1);

      /* Add a X-Linkname header as needed */
      if(options->linkaddr == OPT_DO_LINKADDR)
	{
	  startlen = strlen(hdr);
	  /* Add separator if already a type */
	  if(startlen)
	    {
	      strcpy(hdr + startlen, "\r\n");
	      startlen += 2;
	    }
	  /* Add linkname */
	  sprintf(hdr + startlen, "X-Linkname: %s", extra->linkaddr);
	  http.Headers = hdr;
	}
    }
  else
    {
      /* Create e-Squirt v3 request (as a POST) */
      urllen = strlen(esurl);
      url = malloc(urllen + 1);
      memcpy(url, esurl, urllen + 1);
    }

  /* Post data go there */
  http.PostData = fields->data;
  http.PostLen = fields->datalen;
 
  fprintf(stderr, "URL = [%s] (%d)\n", url, urllen);
  printf("Encoding : [%s]\n", hdr);

  /* Try to fetch the page - local : never use proxy */
  reply = http_request(url, &http, kHMethodPost,
		       HFLAG_RETURN_HEADER | HFLAG_POST_USER_TYPE |
		       HFLAG_FORCE_NO_PROXY);
  /* No longer needed */
  free(url);

  /* Check if we can return some data */
  if(ret_hdr != (gi_obget *) NULL)
    {
      /* Check if TCP did fail */
      if((reply.lSize <= 0) || (reply.pData == NULL))
	{
	  if(reply.pHdr != NULL)
	    free(reply.pHdr);
	  /* Return standard error to server -> will send back NOT_FOUND */
	  return(-1);
	}

      /* Check if HTTP did fail */
      if(strcmp(reply.szHCode, "200"))
	{
	  int	obex_err;

	  /* Cleanup */
	  if(reply.pHdr != NULL)
	    free(reply.pHdr);
	  free(reply.pData);

	  /* Compute Obex error code and return it*/
	  obex_err = ((reply.szHCode[0] - '0') << 4) | (reply.szHCode[2] - '0');
	  printf("Error : [%s] -> %X\n", reply.szHCode, obex_err);
	  return(-obex_err);
	}

      /* Set the name */
      ret_flags->obex.name = OPT_OBEX_NO_NAME;
      /* Set the type */
      ret_flags->obex.type = OPT_OBEX_TYPE_HTML;
      if(reply.pHdr != NULL)
	{
	  //printf("Some headers\n");
	  if(find_header(reply.pHdr, 0, "Content-Type:",
			 extra->type, sizeof(extra->type)))
	    ret_flags->obex.type = OPT_OBEX_TYPE_USER;
	  free(reply.pHdr);
	}

      /* Post process the HTML */

      /* Create space for the data */
      ret_hdr->data = malloc(reply.lSize * 2 + 256);
      if(!ret_hdr->data)
	{
	  free(reply.pData);
	  return(-1);
	}

      /* Mangle the HTML page */
      ret_hdr->datalen = mangle_html(ret_hdr->data, reply.lSize * 2 + 255,
				     reply.pData, reply.lSize,
				     "", "");

      /* Cleanup */
      ret_hdr->cleanup = OPT_DATA_DO_FREE;
      free(reply.pData);

      ret = 1;	/* Reply Header set */
    }
  else
    {
      /* Cleanup */
      if(reply.pHdr != NULL)
	free(reply.pHdr);
      if(reply.pData != NULL)
	free(reply.pData);
      ret = 0;		/* e-Squirt accepted, Reply Header not set */
    }

  /* Return back to the server */
  return(ret);
}

/* ---------------------------------------------------------------- */
/*
 * Receive a Get Interface request.
 */
int
reply_obget(gi_obget *	request,
	    gi_flags *	flags,
	    es_trans *	extra,
	    void *	userdata)
{
  rui_data	html;		/* The RUI string */
  rui_params	params;		/* Processing opts */
  es_squirt	fields;
  es_opt	options;
  char		pagename[ES_MAX_XML];		/* Page name to return */
  char		hostname[ES_MAX_FILENAME];	/* Page name to return */
  char		url[ES_MAX_XML + 16];		/* URL to query */
  char *	hdr = NULL;			/* HTTP header to send */
  int		hdr_len = 0;			/* Length of header */
  HTTP_Method	method = kHMethodGet;		/* Default : GET */
  HTTP_Extra	http;
  HTTP_Response reply;				/* HTTP reply */
  unsigned long	hflags = HFLAG_RETURN_HEADER | HFLAG_POST_USER_TYPE;

  // Binary == yuck...
  debug_obget(request, flags, extra);

  memset(&http, '\0', sizeof(http));

  /* Note : 
   * If it is a simple get-interface, the requester may have set name and
   * type in its request, and we should take this into account.
   * If it is a squirt-get, receive_esquirt() will have already set
   * name and type in the structure */

  /* First : take care of HTTP headers */
  if(flags->obex.http == OPT_OBEX_HAS_HTTP)
    {
      /* Calculate length */
      hdr_len = strlen(extra->http_hdr);
      /* Allocate (including space for type) */
      hdr = malloc(hdr_len + 128);
      memcpy(hdr, extra->http_hdr, hdr_len + 1);
      http.Headers = hdr;
      /* Get rid of trailing \r\n */
      if((hdr_len > 2) && (hdr[hdr_len - 2] == '\r'))
	{
	  hdr[hdr_len - 2] = '\0';
	  hdr_len -= 2;
	}
    }

  /* Ok, now decapsulate the form data... */
  init_structs(&fields, &options, NULL, NULL, NULL, NULL);
  rui_extract_obget(&fields, &options, extra, request, flags,
		    &html, &params);
  debug_rui(&html, &params);
  if(params.form == OPT_RUI_FORM_SQUIRT)
    {
      /* Get rid of potential form data passed to us
       * rui_extract_obget() did copy it to fields */
      free_structs(NULL, extra, request);

      /* Here we process the reference or we call receive_squirt() */
      return(receive_esquirt(&fields, &options, extra, request, flags));
    }
  else
    {
      /* Check for POST forms */
      if(params.form == OPT_RUI_FORM_POST)
	{
	  method = kHMethodPost;
	  /* Check if header exist, add type */
	  if(html.encoding)
	    {
	      if(hdr)
		{
		  strcpy(hdr + hdr_len, "\r\n");
		  hdr_len += 2;
		}
	      else
		{
		  hdr = malloc(128);
		  hdr_len = 0;
		  http.Headers = hdr;
		}
	      sprintf(hdr + hdr_len, "Content-Type: %s", html.encoding);
	    }
	  /* Post data go there */
	  http.PostData = html.input;
	  http.PostLen = html.inputlen;
	}

      /* Check if we have a filename/action/URL */
      if(html.action != NULL)
	{
	  /* May want to do better than that in case of overflow... */
	  if(strlen(html.action) >= ES_MAX_XML)
	    html.action[ES_MAX_XML - 1] = '\0';

	  /* Save it */
	  strcpy(pagename, html.action);
	}
      else
	/* No name */
	strcpy(pagename, "");

      /* Create the proper URL based on the request.
       * If it contains already a ':', it is an absolute URL and
       * we don't need to do anything.
       * If it doesn't contain a ':', it's a relative URL. We need
       * to redirect it to our local web server.
       * Note on ':' : this characted is forbidden in filename and
       * GET arguments (i.e. escaped) and will allow us to catch all
       * variations of "http://" or "ftp://"...
       * Note2 : ':' may be used to specify the port number...
       */
      if(strstr(pagename, ":/"))
	strcpy(url, pagename);
      else
	{
	  /* Relative pages -> local server */
	  strcpy(url, "http://127.0.0.1/");
	  strcat(url, pagename);
	}

      /* Check if we need to disable the proxy (for local URLs) */
      if(!strncasecmp(url, "http://127.0.0.1", 16))
	hflags |= HFLAG_FORCE_NO_PROXY;
    }

  if(hdr)
    printf("HTTP header : [%s]\n", hdr);
  fprintf(stderr, "URL = [%s]\n", url);

  /* Try to fetch the page */
  reply = http_request(url, &http, method, hflags);

  /* Get rid of potential form data passed to us (only after processed it) */
  free_structs(NULL, extra, request);
  if(hdr)
    free(hdr);

  /* Check if TCP did fail */
  if((reply.lSize <= 0) || (reply.pData == NULL))
    {
      if(reply.pHdr != NULL)
	free(reply.pHdr);
      /* Return standard error to server -> will send back NOT_FOUND */
      return(-1);
    }

  /* Check if HTTP did fail */
  if(strcmp(reply.szHCode, "200"))
    {
      int	obex_err;

      /* Cleanup */
      if(reply.pHdr != NULL)
	free(reply.pHdr);
      free(reply.pData);

      /* Compute Obex error code and return it*/
      obex_err = ((reply.szHCode[0] - '0') << 4) | (reply.szHCode[2] - '0');
      printf("Error : [%s] -> %X\n", reply.szHCode, obex_err);
      return(-obex_err);
    }

  /* Set the name */
  strcpy(extra->filename, pagename);
  flags->obex.name = OPT_OBEX_HAS_NAME;
  /* Set the type */
  flags->obex.type = OPT_OBEX_TYPE_HTML;
  if(reply.pHdr != NULL)
    {
      //printf("Some headers [%s]\n", reply.pHdr);
      /* Get type */
      if(find_header(reply.pHdr, 0, "Content-Type:",
		     extra->type, sizeof(extra->type)))
	flags->obex.type = OPT_OBEX_TYPE_USER;
      /* Get other stuff we care about */
      if(http_strip_headers(reply.pHdr))
	{
	  extra->http_hdr = reply.pHdr;
	  extra->http_cleanup = OPT_DATA_DO_FREE;
	  flags->obex.http = OPT_OBEX_HAS_HTTP;
	}
      else
	free(reply.pHdr);
    }

  /* Post process the HTML */

  /* Get base path */
  get_base_path(pagename, hostname);
  //printf("Pagename : [%s] ; Hostname : [%s]\n", pagename, hostname);

  /* Create space for the data */
  request->data = malloc(reply.lSize * 2 + 256);
  if(!request->data)
    {
      free(reply.pData);
      return(-1);
    }

  /* Mangle the HTML page */
  request->datalen = mangle_html(request->data, reply.lSize * 2 + 255,
				 reply.pData, reply.lSize,
				 pagename, hostname);

  /* Cleanup */
  request->cleanup = OPT_DATA_DO_FREE;
  free(reply.pData);
  /* Don't send back the url */
  flags->obex.name = OPT_OBEX_NO_NAME;

  /* Check what we are returning */
  //debug_obget(request, flags, extra);

  /* Return back to the server */
  return(1);
}

/******************************* MAIN *******************************/

/* ---------------------------------------------------------------- */
/*
 * main body of the program
 */
int
main(int	argc,
     char *	argv[])
{
  es_squirt	fields;
  es_opt	options;
  es_trans	extra;
  gi_obget	interface;
  gi_flags	flags;
  es_server_data	control;
  int		ret = -1;

  init_structs(&fields, &options, &extra, &interface, &flags, &control);

  control.acceptback = &server_accept;
  control.timeback = &server_select;
  control.timeout = 0;

  /* Set the receive options that matter */
  options.model = OPT_UNDEFINED;	/* Only for the receiver */
  options.protocol = OPT_UNDEFINED;
  options.do_xml = OPT_NO_XML;	/* Don't handle XML in the lib */
  /* Note : we listen only on IrDA sockets, not Ultra sockets.
   * We might miss on Ultra-e-Squirt, but on the other hand it allow other
   * applications to bind to Ultra and receive Beacons. */
  options.transport = OPT_IRDA | OPT_BLUETOOTH;

  /* Prepare to handle get interface calls */
  set_obget_handler(&reply_obget, NULL);

  /* Prepare to receive signals from our childs */
  set_signal_handlers();

  /* Receive multiple e-Squirt/Beacon */
  printf("Waiting for e-Squirts/RUI over Obex...\n");
  recv_squirt(&fields, &options, &extra, &receive_esquirt, &control);
  /* Will loop forever */

  return(ret);
}
